import argparse
import logging
import sys
from collections.abc import Collection, Mapping
from ipaddress import ip_address
from typing import Any, Final, Optional, cast

from e7awghal import CapIqDataReader

from quel_ic_config import QUEL1_BOXTYPE_ALIAS, Quel1BoxType, Quel1PortType
from quel_ic_config_utils import (
    BoxPool,
    BoxSettingType,
    VportSettingType,
    VportTypicalSettingType,
    complete_ipaddrs,
    find_chunks,
    plot_iqs,
)

logger = logging.getLogger()


DEFAULT_PULSE_DETECTION_THRESHOLD: Final[float] = 2000.0
DEFAULT_PULSE_DETECTION_THRESHOLD_FUJITSU11_MON: Final[float] = 1200.0


CAP_VPORT_SETTINGS_QUEL1SE_FUJITSU11_A: dict[str, dict[str, VportSettingType]] = {
    "READ0": {
        "create": {
            "boxname": "BOX0",
            "port": 0,  # (0, "r")
            "runits": {0},
        },
        "config": {
            "lo_freq": 8.5e9,
            "cnco_freq": 1.5e9,
            "fnco_freq": 0.0,
            "rfswitch": "loop",
        },
        "simple_parameters": {
            0: {
                "num_delay_sample": 0,
                "num_integration_section": 1,
                "num_capture_samples": [2048],
                "num_blank_samples": [4],
            },
        },
        "check": {
            "pulse_threshold": DEFAULT_PULSE_DETECTION_THRESHOLD,
        },
    },
    "READ1": {
        "create": {
            "boxname": "BOX0",
            "port": 7,  # (1, "r")
            "runits": {0},
        },
        "config": {
            "lo_freq": 8.5e9,
            "cnco_freq": 1.5e9,
            "fnco_freq": 0.0,
            "rfswitch": "loop",
        },
        "simple_parameters": {
            0: {
                "num_delay_sample": 0,
                "num_integration_section": 1,
                "num_capture_samples": [2048],
                "num_blank_samples": [4],
            },
        },
        "check": {
            "pulse_threshold": DEFAULT_PULSE_DETECTION_THRESHOLD,
        },
    },
    "MON0": {
        "create": {
            "boxname": "BOX0",
            "port": 5,  # (0, "m")
            "runits": {0},
        },
        "config": {
            "lo_freq": 11.5e9,
            "cnco_freq": 1.5e9,
            "fnco_freq": 0.0,
            "rfswitch": "loop",
        },
        "simple_parameters": {
            0: {
                "num_delay_sample": 0,
                "num_integration_section": 1,
                "num_capture_samples": [2048],
                "num_blank_samples": [4],
            },
        },
        "check": {
            "pulse_threshold": DEFAULT_PULSE_DETECTION_THRESHOLD_FUJITSU11_MON,
        },
    },
    "MON1": {
        "create": {
            "boxname": "BOX0",
            "port": 12,  # (1, "m")
            "runits": {0},
        },
        "config": {
            "lo_freq": 11.5e9,
            "cnco_freq": 1.5e9,
            "fnco_freq": 0.0,
            "rfswitch": "loop",
        },
        "simple_parameters": {
            0: {
                "num_delay_sample": 0,
                "num_integration_section": 1,
                "num_capture_samples": [2048],
                "num_blank_samples": [4],
            },
        },
        "check": {
            "pulse_threshold": DEFAULT_PULSE_DETECTION_THRESHOLD_FUJITSU11_MON,
        },
    },
}

CAP_VPORT_SETTINGS_QUEL1SE_FUJITSU11_B: dict[str, dict[str, VportSettingType]] = {
    "MON0": {
        "create": {
            "boxname": "BOX0",
            "port": 5,  # (0, "m")
            "runits": {0},
        },
        "config": {
            "lo_freq": 11.5e9,
            "cnco_freq": 1.5e9,
            "fnco_freq": 0.0,
            "rfswitch": "loop",
        },
        "simple_parameters": {
            0: {
                "num_delay_sample": 0,
                "num_integration_section": 1,
                "num_capture_samples": [3072],
                "num_blank_samples": [4],
            },
        },
        "check": {
            "pulse_threshold": DEFAULT_PULSE_DETECTION_THRESHOLD_FUJITSU11_MON,
        },
    },
    "MON1": {
        "create": {
            "boxname": "BOX0",
            "port": 12,  # (1, "m")
            "runits": {0},
        },
        "config": {
            "lo_freq": 11.5e9,
            "cnco_freq": 1.5e9,
            "fnco_freq": 0.0,
            "rfswitch": "loop",
        },
        "simple_parameters": {
            0: {
                "num_delay_sample": 0,
                "num_integration_section": 1,
                "num_capture_samples": [3072],
                "num_blank_samples": [4],
            },
        },
        "check": {
            "pulse_threshold": DEFAULT_PULSE_DETECTION_THRESHOLD_FUJITSU11_MON,
        },
    },
}

GEN_VPORT_SETTINGS_QUEL1SE_FUJITSU11_B: dict[str, dict[str, VportSettingType]] = {
    "GEN01": {
        "create": {
            "boxname": "BOX0",
            "port": 1,
            "channel": 0,
        },
        "config": {
            "lo_freq": 11.5e9,
            "cnco_freq": 1.5e9,
            "fnco_freq": 0.0,
            "fullscale_current": 40000,
            "sideband": "L",
            "vatt": 0xC00,
        },
        "cw_parameter": {
            "amplitude": 32767.0,
            "num_wave_sample": 128,
            "num_repeats": (1, 1),
            "num_wait_samples": (0, 0),
        },
    },
    "GEN02": {
        "create": {
            "boxname": "BOX0",
            "port": 2,
            "channel": 0,
        },
        "config": {
            "lo_freq": 11.5e9,
            "cnco_freq": 1.5e9,
            "fnco_freq": 0.0,
            "fullscale_current": 40000,
            "sideband": "L",
            "vatt": 0xC00,
        },
        "cw_parameter": {
            "amplitude": 32767.0,
            "num_wave_sample": 64,
            "num_repeats": (1, 1),
            "num_wait_samples": (512, 0),
        },
    },
    "GEN03": {
        "create": {
            "boxname": "BOX0",
            "port": 3,
            "channel": 0,
        },
        "config": {
            # "lo_freq": 11.5e9,
            "cnco_freq": 1.5e9,
            "fnco_freq": 0.0,
            "fullscale_current": 40000,
            "sideband": "L",
            "vatt": 0xC00,
        },
        "cw_parameter": {
            "amplitude": 32767.0,
            "num_wave_sample": 128,
            "num_repeats": (1, 1),
            "num_wait_samples": (1024, 0),
        },
    },
    "GEN04": {
        "create": {
            "boxname": "BOX0",
            "port": 4,
            "channel": 0,
        },
        "config": {
            "lo_freq": 11.5e9,
            "cnco_freq": 1.5e9,
            "fnco_freq": 0.0,
            "fullscale_current": 40000,
            "sideband": "L",
            "vatt": 0xC00,
        },
        "cw_parameter": {
            "amplitude": 32767.0,
            "num_wave_sample": 64,
            "num_repeats": (1, 1),
            "num_wait_samples": (1536, 0),
        },
    },
    "GEN08": {
        "create": {
            "boxname": "BOX0",
            "port": 8,
            "channel": 0,
        },
        "config": {
            "lo_freq": 11.5e9,
            "cnco_freq": 1.5e9,
            "fnco_freq": 0.0,
            "fullscale_current": 40000,
            "sideband": "L",
            "vatt": 0xC00,
        },
        "cw_parameter": {
            "amplitude": 32767.0,
            "num_wave_sample": 128,
            "num_repeats": (1, 1),
            "num_wait_samples": (0, 0),
        },
    },
    "GEN09": {
        "create": {
            "boxname": "BOX0",
            "port": 9,
            "channel": 0,
        },
        "config": {
            "lo_freq": 11.5e9,
            "cnco_freq": 1.5e9,
            "fnco_freq": 0.0,
            "fullscale_current": 40000,
            "sideband": "L",
            "vatt": 0xC00,
        },
        "cw_parameter": {
            "amplitude": 32767.0,
            "num_wave_sample": 64,
            "num_repeats": (1, 1),
            "num_wait_samples": (512, 0),
        },
    },
    "GEN10": {
        "create": {
            "boxname": "BOX0",
            "port": 10,
            "channel": 0,
        },
        "config": {
            # "lo_freq": 11.5e9,
            "cnco_freq": 1.5e9,
            "fnco_freq": 0.0,
            "fullscale_current": 40000,
            "sideband": "L",
            "vatt": 0xC00,
        },
        "cw_parameter": {
            "amplitude": 32767.0,
            "num_wave_sample": 128,
            "num_repeats": (1, 1),
            "num_wait_samples": (1024, 0),
        },
    },
    "GEN11": {
        "create": {
            "boxname": "BOX0",
            "port": 11,
            "channel": 0,
        },
        "config": {
            "lo_freq": 11.5e9,
            "cnco_freq": 1.5e9,
            "fnco_freq": 0.0,
            "fullscale_current": 40000,
            "sideband": "L",
            "vatt": 0xC00,
        },
        "cw_parameter": {
            "amplitude": 32767.0,
            "num_wave_sample": 64,
            "num_repeats": (1, 1),
            "num_wait_samples": (1536, 0),
        },
    },
}

CAP_VPORT_SETTINGS_QUEL1A: dict[str, dict[str, VportSettingType]] = {
    "READ0": {
        "create": {
            "boxname": "BOX0",
            "port": 0,  # (0, "r")
            "runits": {0},
        },
        "config": {
            "lo_freq": 8.5e9,
            "cnco_freq": 1.5e9,
            "fnco_freq": 0.0,
            "rfswitch": "loop",
        },
        "simple_parameters": {
            0: {
                "num_delay_sample": 0,
                "num_integration_section": 1,
                "num_capture_samples": [2048],
                "num_blank_samples": [4],
            },
        },
        "check": {
            "pulse_threshold": DEFAULT_PULSE_DETECTION_THRESHOLD,
        },
    },
    "READ1": {
        "create": {
            "boxname": "BOX0",
            "port": 7,  # (1, "r")
            "runits": {0},
        },
        "config": {
            "lo_freq": 8.5e9,
            "cnco_freq": 1.5e9,
            "fnco_freq": 0.0,
            "rfswitch": "loop",
        },
        "simple_parameters": {
            0: {
                "num_delay_sample": 0,
                "num_integration_section": 1,
                "num_capture_samples": [2048],
                "num_blank_samples": [4],
            },
        },
        "check": {
            "pulse_threshold": DEFAULT_PULSE_DETECTION_THRESHOLD,
        },
    },
    "MON0": {
        "create": {
            "boxname": "BOX0",
            "port": 5,  # (0, "m")
            "runits": {0},
        },
        "config": {
            "lo_freq": 11.5e9,
            "cnco_freq": 1.5e9,
            "fnco_freq": 0.0,
            "rfswitch": "loop",
        },
        "simple_parameters": {
            0: {
                "num_delay_sample": 0,
                "num_integration_section": 1,
                "num_capture_samples": [2048],
                "num_blank_samples": [4],
            },
        },
        "check": {
            "pulse_threshold": DEFAULT_PULSE_DETECTION_THRESHOLD,
        },
    },
    "MON1": {
        "create": {
            "boxname": "BOX0",
            "port": 12,  # (1, "m")
            "runits": {0},
        },
        "config": {
            "lo_freq": 11.5e9,
            "cnco_freq": 1.5e9,
            "fnco_freq": 0.0,
            "rfswitch": "loop",
        },
        "simple_parameters": {
            0: {
                "num_delay_sample": 0,
                "num_integration_section": 1,
                "num_capture_samples": [2048],
                "num_blank_samples": [4],
            },
        },
        "check": {
            "pulse_threshold": DEFAULT_PULSE_DETECTION_THRESHOLD,
        },
    },
}

GEN_VPORT_SETTINGS_QUEL1A: dict[str, dict[str, VportSettingType]] = {
    "GEN01": {
        "create": {
            "boxname": "BOX0",
            "port": 1,
            "channel": 0,
        },
        "config": {
            # "lo_freq": 8.5e9,
            "cnco_freq": 1.5e9,
            "fnco_freq": 0.0,
            "fullscale_current": 40000,
            "sideband": "U",
            "vatt": 0xC00,
        },
        "cw_parameter": {
            "amplitude": 32767.0,
            "num_wave_sample": 64,
            "num_repeats": (1, 1),
            "num_wait_samples": (0, 0),
        },
    },
    "GEN02": {
        "create": {
            "boxname": "BOX0",
            "port": 2,  # ctrl
            "channel": 0,
        },
        "config": {
            "lo_freq": 11.5e9,
            "cnco_freq": 1.5e9,
            "fnco_freq": 0.0,
            "fullscale_current": 40000,
            "sideband": "L",
            "vatt": 0xC00,
        },
        "cw_parameter": {
            "amplitude": 32767.0,
            "num_wave_sample": 64,
            "num_repeats": (1, 1),
            "num_wait_samples": (0, 0),
        },
    },
    "GEN03": {
        "create": {
            "boxname": "BOX0",
            "port": 3,  # pump-#0
            "channel": 0,
        },
        "config": {
            # "lo_freq": 11.5e9,
            "cnco_freq": 1.5e9,
            "fnco_freq": 0.0,
            "fullscale_current": 40000,
            "sideband": "L",
            "vatt": 0xC00,
        },
        "cw_parameter": {
            "amplitude": 32767.0,
            "num_wave_sample": 128,
            "num_repeats": (1, 1),
            "num_wait_samples": (512, 0),
        },
    },
    "GEN04": {
        "create": {
            "boxname": "BOX0",
            "port": 4,  # ctrl
            "channel": 0,
        },
        "config": {
            "lo_freq": 11.5e9,
            "cnco_freq": 1.5e9,
            "fnco_freq": 0.0,
            "fullscale_current": 40000,
            "sideband": "L",
            "vatt": 0xC00,
        },
        "cw_parameter": {
            "amplitude": 32767.0,
            "num_wave_sample": 64,
            "num_repeats": (1, 1),
            "num_wait_samples": (1024, 0),
        },
    },
    "GEN08": {
        "create": {
            "boxname": "BOX0",
            "port": 8,  # readout
            "channel": 0,
        },
        "config": {
            # "lo_freq": 8.5e9,
            "cnco_freq": 1.5e9,
            "fnco_freq": 0.0,
            "fullscale_current": 40000,
            "sideband": "U",
            "vatt": 0xC00,
        },
        "cw_parameter": {
            "amplitude": 32767.0,
            "num_wave_sample": 64,
            "num_repeats": (1, 1),
            "num_wait_samples": (0, 0),
        },
    },
    "GEN09": {
        "create": {
            "boxname": "BOX0",
            "port": 9,  # ctrl
            "channel": 0,
        },
        "config": {
            "lo_freq": 11.5e9,
            "cnco_freq": 1.5e9,
            "fnco_freq": 0.0,
            "fullscale_current": 40000,
            "sideband": "L",
            "vatt": 0xC00,
        },
        "cw_parameter": {
            "amplitude": 32767.0,
            "num_wave_sample": 64,
            "num_repeats": (1, 1),
            "num_wait_samples": (0, 0),
        },
    },
    "GEN10": {
        "create": {
            "boxname": "BOX0",
            "port": 10,  # pump
            "channel": 0,
        },
        "config": {
            # "lo_freq": 11.5e9,
            "cnco_freq": 1.5e9,
            "fnco_freq": 0.0,
            "fullscale_current": 40000,
            "sideband": "L",
            "vatt": 0xC00,
        },
        "cw_parameter": {
            "amplitude": 32767.0,
            "num_wave_sample": 128,
            "num_repeats": (1, 1),
            "num_wait_samples": (512, 0),
        },
    },
    "GEN11": {
        "create": {
            "boxname": "BOX0",
            "port": 11,  # ctrl
            "channel": 0,
        },
        "config": {
            "lo_freq": 11.5e9,
            "cnco_freq": 1.5e9,
            "fnco_freq": 0.0,
            "fullscale_current": 40000,
            "sideband": "L",
            "vatt": 0xC00,
        },
        "cw_parameter": {
            "amplitude": 32767.0,
            "num_wave_sample": 64,
            "num_repeats": (1, 1),
            "num_wait_samples": (1024, 0),
        },
    },
}

CAP_VPORT_SETTINGS_QUEL1B: dict[str, dict[str, VportSettingType]] = {
    "MON0": {
        "create": {
            "boxname": "BOX0",
            "port": 5,  # (0, "m")
            "runits": {0},
        },
        "config": {
            "lo_freq": 11.5e9,
            "cnco_freq": 1.5e9,
            "fnco_freq": 0.0,
            "rfswitch": "loop",
        },
        "simple_parameters": {
            0: {
                "num_delay_sample": 0,
                "num_integration_section": 1,
                "num_capture_samples": [3072],
                "num_blank_samples": [4],
            },
        },
        "check": {
            "pulse_threshold": DEFAULT_PULSE_DETECTION_THRESHOLD,
        },
    },
    "MON1": {
        "create": {
            "boxname": "BOX0",
            "port": 12,  # (1, "m")
            "runits": {0},
        },
        "config": {
            "lo_freq": 11.5e9,
            "cnco_freq": 1.5e9,
            "fnco_freq": 0.0,
            "rfswitch": "loop",
        },
        "simple_parameters": {
            0: {
                "num_delay_sample": 0,
                "num_integration_section": 1,
                "num_capture_samples": [3072],
                "num_blank_samples": [4],
            },
        },
        "check": {
            "pulse_threshold": DEFAULT_PULSE_DETECTION_THRESHOLD,
        },
    },
}

GEN_VPORT_SETTINGS_QUEL1B: dict[str, dict[str, VportSettingType]] = {
    "GEN01": {
        "create": {
            "boxname": "BOX0",
            "port": 1,
            "channel": 0,
        },
        "config": {
            "lo_freq": 11.5e9,
            "cnco_freq": 1.5e9,
            "fnco_freq": 0.0,
            "fullscale_current": 40000,
            "sideband": "L",
            "vatt": 0xC00,
        },
        "cw_parameter": {
            "amplitude": 32767.0,
            "num_wave_sample": 128,
            "num_repeats": (1, 1),
            "num_wait_samples": (0, 0),
        },
    },
    "GEN02": {
        "create": {
            "boxname": "BOX0",
            "port": 2,
            "channel": 0,
        },
        "config": {
            # "lo_freq": 11.5e9,
            "cnco_freq": 1.5e9,
            "fnco_freq": 0.0,
            "fullscale_current": 40000,
            "sideband": "L",
            "vatt": 0xC00,
        },
        "cw_parameter": {
            "amplitude": 32767.0,
            "num_wave_sample": 64,
            "num_repeats": (1, 1),
            "num_wait_samples": (512, 0),
        },
    },
    "GEN03": {
        "create": {
            "boxname": "BOX0",
            "port": 3,
            "channel": 0,
        },
        "config": {
            "lo_freq": 11.5e9,
            "cnco_freq": 1.5e9,
            "fnco_freq": 0.0,
            "fullscale_current": 40000,
            "sideband": "L",
            "vatt": 0xC00,
        },
        "cw_parameter": {
            "amplitude": 32767.0,
            "num_wave_sample": 128,
            "num_repeats": (1, 1),
            "num_wait_samples": (1024, 0),
        },
    },
    "GEN04": {
        "create": {
            "boxname": "BOX0",
            "port": 4,
            "channel": 0,
        },
        "config": {
            "lo_freq": 11.5e9,
            "cnco_freq": 1.5e9,
            "fnco_freq": 0.0,
            "fullscale_current": 40000,
            "sideband": "L",
            "vatt": 0xC00,
        },
        "cw_parameter": {
            "amplitude": 32767.0,
            "num_wave_sample": 64,
            "num_repeats": (1, 1),
            "num_wait_samples": (1536, 0),
        },
    },
    "GEN08": {
        "create": {
            "boxname": "BOX0",
            "port": 8,
            "channel": 0,
        },
        "config": {
            "lo_freq": 11.5e9,
            "cnco_freq": 1.5e9,
            "fnco_freq": 0.0,
            "fullscale_current": 40000,
            "sideband": "L",
            "vatt": 0xC00,
        },
        "cw_parameter": {
            "amplitude": 32767.0,
            "num_wave_sample": 128,
            "num_repeats": (1, 1),
            "num_wait_samples": (0, 0),
        },
    },
    "GEN09": {
        "create": {
            "boxname": "BOX0",
            "port": 9,
            "channel": 0,
        },
        "config": {
            # "lo_freq": 11.5e9,
            "cnco_freq": 1.5e9,
            "fnco_freq": 0.0,
            "fullscale_current": 40000,
            "sideband": "L",
            "vatt": 0xC00,
        },
        "cw_parameter": {
            "amplitude": 32767.0,
            "num_wave_sample": 64,
            "num_repeats": (1, 1),
            "num_wait_samples": (512, 0),
        },
    },
    "GEN10": {
        "create": {
            "boxname": "BOX0",
            "port": 10,
            "channel": 0,
        },
        "config": {
            "lo_freq": 11.5e9,
            "cnco_freq": 1.5e9,
            "fnco_freq": 0.0,
            "fullscale_current": 40000,
            "sideband": "L",
            "vatt": 0xC00,
        },
        "cw_parameter": {
            "amplitude": 32767.0,
            "num_wave_sample": 128,
            "num_repeats": (1, 1),
            "num_wait_samples": (1024, 0),
        },
    },
    "GEN11": {
        "create": {
            "boxname": "BOX0",
            "port": 11,
            "channel": 0,
        },
        "config": {
            "lo_freq": 11.5e9,
            "cnco_freq": 1.5e9,
            "fnco_freq": 0.0,
            "fullscale_current": 40000,
            "sideband": "L",
            "vatt": 0xC00,
        },
        "cw_parameter": {
            "amplitude": 32767.0,
            "num_wave_sample": 64,
            "num_repeats": (1, 1),
            "num_wait_samples": (1536, 0),
        },
    },
}

CAP_VPORT_SETTINGS_QUBERIKENA: dict[str, dict[str, VportSettingType]] = {
    "READ0": {
        "create": {
            "boxname": "BOX0",
            "port": 1,  # (0, "r")
            "runits": {0},
        },
        "config": {
            "lo_freq": 8.5e9,
            "cnco_freq": 1.5e9,
            "fnco_freq": 0.0,
            "rfswitch": "loop",
        },
        "simple_parameters": {
            0: {
                "num_delay_sample": 0,
                "num_integration_section": 1,
                "num_capture_samples": [2048],
                "num_blank_samples": [4],
            },
        },
        "check": {
            "pulse_threshold": DEFAULT_PULSE_DETECTION_THRESHOLD,
        },
    },
    "READ1": {
        "create": {
            "boxname": "BOX0",
            "port": 12,  # (1, "r")
            "runits": {0},
        },
        "config": {
            "lo_freq": 8.5e9,
            "cnco_freq": 1.5e9,
            "fnco_freq": 0.0,
            "rfswitch": "loop",
        },
        "simple_parameters": {
            0: {
                "num_delay_sample": 0,
                "num_integration_section": 1,
                "num_capture_samples": [2048],
                "num_blank_samples": [4],
            },
        },
        "check": {
            "pulse_threshold": DEFAULT_PULSE_DETECTION_THRESHOLD,
        },
    },
    "MON0": {
        "create": {
            "boxname": "BOX0",
            "port": 4,  # (0, "m")
            "runits": {0},
        },
        "config": {
            "lo_freq": 11.5e9,
            "cnco_freq": 1.5e9,
            "fnco_freq": 0.0,
            "rfswitch": "loop",
        },
        "simple_parameters": {
            0: {
                "num_delay_sample": 0,
                "num_integration_section": 1,
                "num_capture_samples": [2048],
                "num_blank_samples": [4],
            },
        },
        "check": {
            "pulse_threshold": DEFAULT_PULSE_DETECTION_THRESHOLD,
        },
    },
    "MON1": {
        "create": {
            "boxname": "BOX0",
            "port": 9,  # (1, "m")
            "runits": {0},
        },
        "config": {
            "lo_freq": 11.5e9,
            "cnco_freq": 1.5e9,
            "fnco_freq": 0.0,
            "rfswitch": "loop",
        },
        "simple_parameters": {
            0: {
                "num_delay_sample": 0,
                "num_integration_section": 1,
                "num_capture_samples": [2048],
                "num_blank_samples": [4],
            },
        },
        "check": {
            "pulse_threshold": DEFAULT_PULSE_DETECTION_THRESHOLD,
        },
    },
}

GEN_VPORT_SETTINGS_QUBERIKENA: dict[str, dict[str, VportSettingType]] = {
    "GEN00": {
        "create": {
            "boxname": "BOX0",
            "port": 0,  # readout
            "channel": 0,
        },
        "config": {
            # "lo_freq": 8.5e9,
            "cnco_freq": 1.5e9,
            "fnco_freq": 0.0,
            "fullscale_current": 40000,
            "sideband": "U",
            "vatt": 0xC00,
        },
        "cw_parameter": {
            "amplitude": 32767.0,
            "num_wave_sample": 64,
            "num_repeats": (1, 1),
            "num_wait_samples": (0, 0),
        },
    },
    "GEN02": {
        "create": {
            "boxname": "BOX0",
            "port": 2,  # pump
            "channel": 0,
        },
        "config": {
            # "lo_freq": 11.5e9,
            "cnco_freq": 1.5e9,
            "fnco_freq": 0.0,
            "fullscale_current": 40000,
            "sideband": "L",
            "vatt": 0xC00,
        },
        "cw_parameter": {
            "amplitude": 32767.0,
            "num_wave_sample": 64,
            "num_repeats": (1, 1),
            "num_wait_samples": (0, 0),
        },
    },
    "GEN05": {
        "create": {
            "boxname": "BOX0",
            "port": 5,  # ctrl
            "channel": 0,
        },
        "config": {
            "lo_freq": 11.5e9,
            "cnco_freq": 1.5e9,
            "fnco_freq": 0.0,
            "fullscale_current": 40000,
            "sideband": "L",
            "vatt": 0xC00,
        },
        "cw_parameter": {
            "amplitude": 32767.0,
            "num_wave_sample": 128,
            "num_repeats": (1, 1),
            "num_wait_samples": (512, 0),
        },
    },
    "GEN06": {
        "create": {
            "boxname": "BOX0",
            "port": 6,  # ctrl
            "channel": 0,
        },
        "config": {
            "lo_freq": 11.5e9,
            "cnco_freq": 1.5e9,
            "fnco_freq": 0.0,
            "fullscale_current": 40000,
            "sideband": "L",
            "vatt": 0xC00,
        },
        "cw_parameter": {
            "amplitude": 32767.0,
            "num_wave_sample": 64,
            "num_repeats": (1, 1),
            "num_wait_samples": (1024, 0),
        },
    },
    "GEN13": {
        "create": {
            "boxname": "BOX0",
            "port": 13,  # readout
            "channel": 0,
        },
        "config": {
            # "lo_freq": 8.5e9,
            "cnco_freq": 1.5e9,
            "fnco_freq": 0.0,
            "fullscale_current": 40000,
            "sideband": "U",
            "vatt": 0xC00,
        },
        "cw_parameter": {
            "amplitude": 32767.0,
            "num_wave_sample": 64,
            "num_repeats": (1, 1),
            "num_wait_samples": (0, 0),
        },
    },
    "GEN11": {
        "create": {
            "boxname": "BOX0",
            "port": 11,  # pump
            "channel": 0,
        },
        "config": {
            # "lo_freq": 11.5e9,
            "cnco_freq": 1.5e9,
            "fnco_freq": 0.0,
            "fullscale_current": 40000,
            "sideband": "L",
            "vatt": 0xC00,
        },
        "cw_parameter": {
            "amplitude": 32767.0,
            "num_wave_sample": 64,
            "num_repeats": (1, 1),
            "num_wait_samples": (0, 0),
        },
    },
    "GEN08": {
        "create": {
            "boxname": "BOX0",
            "port": 8,  # ctrl
            "channel": 0,
        },
        "config": {
            "lo_freq": 11.5e9,
            "cnco_freq": 1.5e9,
            "fnco_freq": 0.0,
            "fullscale_current": 40000,
            "sideband": "L",
            "vatt": 0xC00,
        },
        "cw_parameter": {
            "amplitude": 32767.0,
            "num_wave_sample": 128,
            "num_repeats": (1, 1),
            "num_wait_samples": (512, 0),
        },
    },
    "GEN07": {
        "create": {
            "boxname": "BOX0",
            "port": 7,  # ctrl
            "channel": 0,
        },
        "config": {
            "lo_freq": 11.5e9,
            "cnco_freq": 1.5e9,
            "fnco_freq": 0.0,
            "fullscale_current": 40000,
            "sideband": "L",
            "vatt": 0xC00,
        },
        "cw_parameter": {
            "amplitude": 32767.0,
            "num_wave_sample": 64,
            "num_repeats": (1, 1),
            "num_wait_samples": (1024, 0),
        },
    },
}

CAP_VPORT_SETTINGS_QUBERIKENB: dict[str, dict[str, VportSettingType]] = {
    "MON0": {
        "create": {
            "boxname": "BOX0",
            "port": 4,  # (0, "m")
            "runits": {0},
        },
        "config": {
            "lo_freq": 11.5e9,
            "cnco_freq": 1.5e9,
            "fnco_freq": 0.0,
            "rfswitch": "loop",
        },
        "simple_parameters": {
            0: {
                "num_delay_sample": 0,
                "num_integration_section": 1,
                "num_capture_samples": [3072],
                "num_blank_samples": [4],
            },
        },
        "check": {
            "pulse_threshold": DEFAULT_PULSE_DETECTION_THRESHOLD,
        },
    },
    "MON1": {
        "create": {
            "boxname": "BOX0",
            "port": 9,  # (1, "m")
            "runits": {0},
        },
        "config": {
            "lo_freq": 11.5e9,
            "cnco_freq": 1.5e9,
            "fnco_freq": 0.0,
            "rfswitch": "loop",
        },
        "simple_parameters": {
            0: {
                "num_delay_sample": 0,
                "num_integration_section": 1,
                "num_capture_samples": [3072],
                "num_blank_samples": [4],
            },
        },
        "check": {
            "pulse_threshold": DEFAULT_PULSE_DETECTION_THRESHOLD,
        },
    },
}

GEN_VPORT_SETTINGS_QUBERIKENB: dict[str, dict[str, VportSettingType]] = {
    "GEN00": {
        "create": {
            "boxname": "BOX0",
            "port": 0,
            "channel": 0,
        },
        "config": {
            "lo_freq": 11.5e9,
            "cnco_freq": 1.5e9,
            "fnco_freq": 0.0,
            "fullscale_current": 40000,
            "sideband": "L",
            "vatt": 0xC00,
        },
        "cw_parameter": {
            "amplitude": 32767.0,
            "num_wave_sample": 128,
            "num_repeats": (1, 1),
            "num_wait_samples": (0, 0),
        },
    },
    "GEN02": {
        "create": {
            "boxname": "BOX0",
            "port": 2,
            "channel": 0,
        },
        "config": {
            # "lo_freq": 11.5e9,
            "cnco_freq": 1.5e9,
            "fnco_freq": 0.0,
            "fullscale_current": 40000,
            "sideband": "L",
            "vatt": 0xC00,
        },
        "cw_parameter": {
            "amplitude": 32767.0,
            "num_wave_sample": 64,
            "num_repeats": (1, 1),
            "num_wait_samples": (512, 0),
        },
    },
    "GEN05": {
        "create": {
            "boxname": "BOX0",
            "port": 5,
            "channel": 0,
        },
        "config": {
            "lo_freq": 11.5e9,
            "cnco_freq": 1.5e9,
            "fnco_freq": 0.0,
            "fullscale_current": 40000,
            "sideband": "L",
            "vatt": 0xC00,
        },
        "cw_parameter": {
            "amplitude": 32767.0,
            "num_wave_sample": 128,
            "num_repeats": (1, 1),
            "num_wait_samples": (1024, 0),
        },
    },
    "GEN06": {
        "create": {
            "boxname": "BOX0",
            "port": 6,
            "channel": 0,
        },
        "config": {
            "lo_freq": 11.5e9,
            "cnco_freq": 1.5e9,
            "fnco_freq": 0.0,
            "fullscale_current": 40000,
            "sideband": "L",
            "vatt": 0xC00,
        },
        "cw_parameter": {
            "amplitude": 32767.0,
            "num_wave_sample": 64,
            "num_repeats": (1, 1),
            "num_wait_samples": (1536, 0),
        },
    },
    "GEN13": {
        "create": {
            "boxname": "BOX0",
            "port": 13,
            "channel": 0,
        },
        "config": {
            "lo_freq": 11.5e9,
            "cnco_freq": 1.5e9,
            "fnco_freq": 0.0,
            "fullscale_current": 40000,
            "sideband": "L",
            "vatt": 0xC00,
        },
        "cw_parameter": {
            "amplitude": 32767.0,
            "num_wave_sample": 128,
            "num_repeats": (1, 1),
            "num_wait_samples": (0, 0),
        },
    },
    "GEN11": {
        "create": {
            "boxname": "BOX0",
            "port": 11,
            "channel": 0,
        },
        "config": {
            # "lo_freq": 11.5e9,
            "cnco_freq": 1.5e9,
            "fnco_freq": 0.0,
            "fullscale_current": 40000,
            "sideband": "L",
            "vatt": 0xC00,
        },
        "cw_parameter": {
            "amplitude": 32767.0,
            "num_wave_sample": 64,
            "num_repeats": (1, 1),
            "num_wait_samples": (512, 0),
        },
    },
    "GEN08": {
        "create": {
            "boxname": "BOX0",
            "port": 8,
            "channel": 0,
        },
        "config": {
            "lo_freq": 11.5e9,
            "cnco_freq": 1.5e9,
            "fnco_freq": 0.0,
            "fullscale_current": 40000,
            "sideband": "L",
            "vatt": 0xC00,
        },
        "cw_parameter": {
            "amplitude": 32767.0,
            "num_wave_sample": 128,
            "num_repeats": (1, 1),
            "num_wait_samples": (1024, 0),
        },
    },
    "GEN07": {
        "create": {
            "boxname": "BOX0",
            "port": 7,
            "channel": 0,
        },
        "config": {
            "lo_freq": 11.5e9,
            "cnco_freq": 1.5e9,
            "fnco_freq": 0.0,
            "fullscale_current": 40000,
            "sideband": "L",
            "vatt": 0xC00,
        },
        "cw_parameter": {
            "amplitude": 32767.0,
            "num_wave_sample": 64,
            "num_repeats": (1, 1),
            "num_wait_samples": (1536, 0),
        },
    },
}


def parse_boxtype(boxtypename: str) -> Quel1BoxType:
    if boxtypename not in QUEL1_BOXTYPE_ALIAS:
        raise ValueError
    return Quel1BoxType.fromstr(boxtypename)


class LoopbackTest:
    def __init__(
        self,
        cm_setting: Mapping[str, Any],
        box_settings: Mapping[str, BoxSettingType],
        cap_vport_settings: Mapping[str, Mapping[str, VportSettingType]],
        gen_vport_settings: Mapping[str, Mapping[str, VportSettingType]],
        cap_gen_map: Mapping[str, tuple[int, ...]],
    ):
        self.cap_vport_settings: Mapping[str, Mapping[str, VportSettingType]] = cap_vport_settings
        self.gen_vport_settings: Mapping[str, Mapping[str, VportSettingType]] = gen_vport_settings
        self.cap_gen_map: Mapping[str, tuple[int, ...]] = cap_gen_map
        self.boxpool = BoxPool(cm_setting, box_settings, cap_vport_settings, gen_vport_settings)
        self.cps = set(self.cap_vport_settings.keys())
        self.pgs = set(self.gen_vport_settings.keys())

    def initialize(self):
        self.boxpool.initialize(allow_resync=False)

    def _do_background_check(self, pulse_threshold: Optional[float] = None) -> None:
        bgnoise = self.boxpool.check_background_noise(self.cps)
        for cpname, bgnoise_max in bgnoise.items():
            check_setting = cast(VportTypicalSettingType, self.cap_vport_settings[cpname]["check"])
            thr: float = pulse_threshold or check_setting["pulse_threshold"]
            if bgnoise_max > thr * 0.75:
                logger.warning(
                    f"the runit '{cpname}' is too noisy for the given power threshold of pulse detection (= {thr})"
                    "you may see spurious pulses in the results"
                )

    def _single_schedule(self, cp: str, pgs: Collection[str], power_thr: float):
        _, c_tasks, g_tasks = self.boxpool.start_at(runit_name=cp, channel_names=pgs, min_time_offset=125_000_000 // 10)
        for boxname, g_task in g_tasks.items():
            g_task.result()
            logger.info(f"wave generation of box {boxname} is completed")

        rdrs: dict[str, dict[tuple[Quel1PortType, int], CapIqDataReader]] = {}
        for boxname, c_task in c_tasks.items():
            rdrs[boxname] = c_task.result()
            logger.info(f"capture of box {boxname} is completed")

        if len(rdrs) != 1:
            raise AssertionError("too much reader objects...")

        cp_box, cp_port, cp_runits = self.boxpool._runits[cp]

        rdr = rdrs[cp_box]
        iq = rdr[cp_port, list(cp_runits)[0]].as_wave_dict()["s0"][0]
        chunks = find_chunks(iq, power_thr=power_thr)
        return iq, chunks

    def _test_loopback(self, cp: str, pgs: Collection[str], power_thr: float):
        # check_background_noise(cp, power_thr)
        iq, chunks = self._single_schedule(cp, pgs, power_thr)
        if len(chunks) != len(pgs):
            logger.error(
                f"the number of pulses captured by the runit '{cp}` is "
                f"expected to be {len(pgs)} but is actually {len(chunks)}, something wrong"
            )
        return iq, chunks

    def do_test(self, pulse_detection_threshold: Optional[float] = None):
        self._do_background_check()

        # Notes: it is fine to measure time diff with a single input port because all the input ports belong to the same
        #        box in this script.
        self.boxpool.measure_timediff(list(self.cps)[0])

        epochs = {}
        for capname, genidxs in self.cap_gen_map.items():
            gennames = [f"GEN{idx:02d}" for idx in genidxs]
            thr: float = cast(VportTypicalSettingType, self.cap_vport_settings[capname]["check"])["pulse_threshold"]
            iqs, chunks = self._test_loopback(
                capname,
                gennames,
                cast(float, pulse_detection_threshold or thr),
            )

            ll = [f"{capname}: "]
            ll.extend(gennames)
            label = " ".join(ll)
            epochs[label] = iqs

        plot_iqs(epochs, same_range=False)


TEST_CONFIGS = {
    Quel1BoxType.QuEL1SE_FUJITSU11_TypeA: lambda cms, bxs: LoopbackTest(
        cms,
        bxs,
        CAP_VPORT_SETTINGS_QUEL1SE_FUJITSU11_A,
        GEN_VPORT_SETTINGS_QUEL1A,
        {
            "READ0": (1,),
            "MON0": (2, 3, 4),
            "READ1": (8,),
            "MON1": (9, 10, 11),
        },
    ),
    Quel1BoxType.QuEL1SE_FUJITSU11_TypeB: lambda cms, bxs: LoopbackTest(
        cms,
        bxs,
        CAP_VPORT_SETTINGS_QUEL1SE_FUJITSU11_B,
        GEN_VPORT_SETTINGS_QUEL1SE_FUJITSU11_B,
        {
            "MON0": (1, 2, 3, 4),
            "MON1": (8, 9, 10, 11),
        },
    ),
    Quel1BoxType.QuEL1_TypeA: lambda cms, bxs: LoopbackTest(
        cms,
        bxs,
        CAP_VPORT_SETTINGS_QUEL1A,
        GEN_VPORT_SETTINGS_QUEL1A,
        {
            "READ0": (1,),
            "MON0": (2, 3, 4),
            "READ1": (8,),
            "MON1": (9, 10, 11),
        },
    ),
    Quel1BoxType.QuEL1_TypeB: lambda cms, bxs: LoopbackTest(
        cms,
        bxs,
        CAP_VPORT_SETTINGS_QUEL1B,
        GEN_VPORT_SETTINGS_QUEL1B,
        {
            "MON0": (1, 2, 3, 4),
            "MON1": (8, 9, 10, 11),
        },
    ),
    Quel1BoxType.QuBE_RIKEN_TypeA: lambda cms, bxs: LoopbackTest(
        cms,
        bxs,
        CAP_VPORT_SETTINGS_QUBERIKENA,
        GEN_VPORT_SETTINGS_QUBERIKENA,
        {
            "READ0": (0,),
            "MON0": (2, 5, 6),
            "READ1": (13,),
            "MON1": (11, 8, 7),
        },
    ),
    Quel1BoxType.QuBE_RIKEN_TypeB: lambda cms, bxs: LoopbackTest(
        cms,
        bxs,
        CAP_VPORT_SETTINGS_QUBERIKENB,
        GEN_VPORT_SETTINGS_QUBERIKENB,
        {
            "MON0": (0, 2, 5, 6),
            "MON1": (13, 11, 8, 7),
        },
    ),
}


def main():
    logging.basicConfig(level=logging.INFO, format="{asctime} [{levelname:.4}] {name}: {message}", style="{")
    for lgrname, lgr in logging.root.manager.loggerDict.items():
        if lgrname in {"root"}:
            pass
        elif lgrname.startswith("quel_ic_config_utils.simple_multibox_framework"):
            pass
        else:
            if isinstance(lgr, logging.Logger):
                lgr.setLevel(logging.WARNING)

    parser = argparse.ArgumentParser("observing signals from all the output port via internal loop-back paths")
    parser.add_argument(
        "--ipaddr_wss",
        type=ip_address,
        required=True,
        help="IP address of the wave generation/capture subsystem of the target box",
    )
    parser.add_argument(
        "--ipaddr_sss",
        type=ip_address,
        default=0,
        help="IP address of the synchronization subsystem of the target box",
    )
    parser.add_argument(
        "--ipaddr_css",
        type=ip_address,
        default=0,
        help="IP address of the configuration subsystem of the target box",
    )
    parser.add_argument(
        "--boxtype",
        type=parse_boxtype,
        required=True,
        help=f"a type of the target box A: either of "
        f"{', '.join([t for t in QUEL1_BOXTYPE_ALIAS if not t.startswith('x_')])}",
    )
    parser.add_argument(
        "--pulse_detection_threshold",
        type=float,
        default=None,
        help="threshold of amplitude for detecting the pulses",
    )

    args = parser.parse_args()
    complete_ipaddrs(args)

    if args.boxtype not in TEST_CONFIGS:
        logger.error(f"boxtype '{Quel1BoxType.tostr(args.boxtype)}' is not supported")
        sys.exit(1)

    CLOCKMASTER_SETTINGS: dict[str, Any] = {}

    BOX_SETTINGS: dict[str, BoxSettingType] = {
        "BOX0": {
            "ipaddr_wss": str(args.ipaddr_wss),
            "ipaddr_sss": str(args.ipaddr_sss),
            "ipaddr_css": str(args.ipaddr_css),
            "boxtype": args.boxtype,
            "ignore_crc_error_of_mxfe": {0, 1},
        },
    }

    test_config = TEST_CONFIGS[args.boxtype](CLOCKMASTER_SETTINGS, BOX_SETTINGS)
    test_config.initialize()
    test_config.do_test(pulse_detection_threshold=args.pulse_detection_threshold)

    del test_config


if __name__ == "__main__":
    main()



from e7awghal import AwgParam, CapIqDataReader, CapParam, CapSection, E7FwType, WaveChunk

from quel_ic_config.ad5328 import Ad5328ConfigHelper, Ad5328RegNames, Ad5328Regs
from quel_ic_config.ad9082 import Ad9082JesdParam, ChipTemperatures, LinkStatus, NcoFtw
from quel_ic_config.adrf6780 import Adrf6780ConfigHelper, Adrf6780LoSideband, Adrf6780RegNames, Adrf6780Regs
from quel_ic_config.box_force_unlock import force_unlock_all_boxes
from quel_ic_config.box_lock import BoxLockError, set_trancated_traceback_for_lock_error
from quel_ic_config.e7resource_mapper import AbstractQuel1E7ResourceMapper, Quel1ConventionalE7ResourceMapper
from quel_ic_config.exstickge_coap_client import Quel1seBoard, get_exstickge_server_info
from quel_ic_config.linkupper import LinkupFpgaMxfe, LinkupStatistic, LinkupStatus
from quel_ic_config.lmx2594 import Lmx2594ConfigHelper, Lmx2594RegNames, Lmx2594Regs
from quel_ic_config.mixerboard_gpio import MixerboardGpioConfigHelper, MixerboardGpioRegNames, MixerboardGpioRegs
from quel_ic_config.pathselectorboard_gpio import (
    PathselectorboardGpioConfigHelper,
    PathselectorboardGpioRegNames,
    PathselectorboardGpioRegs,
)
from quel_ic_config.powerboard_pwm import PowerboardPwmConfigHelper, PowerboardPwmRegs, PowerboardPwmRegsName
from quel_ic_config.quel1_any_config_subsystem import Quel1AnyConfigSubsystem, Quel1seAnyConfigSubsystem
from quel_ic_config.quel1_box import (
    BoxStartCapunitsByTriggerTask,
    BoxStartCapunitsNowTask,
    Quel1Box,
    Quel1PortType,
    parse_port_str,
)
from quel_ic_config.quel1_box_intrinsic import (
    BoxIntrinsicStartCapunitsByTriggerTask,
    BoxIntrinsicStartCapunitsNowTask,
    Quel1BoxIntrinsic,
    Quel1LineType,
)
from quel_ic_config.quel1_config_loader import Quel1ConfigLoader
from quel_ic_config.quel1_config_subsystem import (
    Ad9082Quel1,
    ExstickgeSockClientQuel1WithDummyLock,
    QubeConfigSubsystem,
    QubeOuTypeAConfigSubsystem,
    QubeOuTypeBConfigSubsystem,
    Quel1ConfigSubsystem,
    Quel1NecConfigSubsystem,
    Quel1TypeAConfigSubsystem,
    Quel1TypeBConfigSubsystem,
)
from quel_ic_config.quel1_config_subsystem_common import NoLoopbackPathError, NoRfSwitchError, Quel1ConfigSubsystemRoot
from quel_ic_config.quel1_config_subsystem_tempctrl import Quel1seTempctrlState
from quel_ic_config.quel1_thermistor import (
    Quel1NormalThermistor,
    Quel1PathSelectorThermistor,
    Quel1seExternalThermistor,
    Quel1seOnboardThermistor,
    Quel1Thermistor,
)
from quel_ic_config.quel1_wave_subsystem import (
    AbstractStartAwgunitsTask,
    Quel1WaveSubsystem,
    StartAwgunitsNowTask,
    StartAwgunitsTimedTask,
    StartCapunitsByTriggerTask,
    StartCapunitsNowTask,
)
from quel_ic_config.quel1se_adda_config_subsystem import (
    ExstickgeCoapClientAdda,
    Quel1seAddaConfigSubsystem,
    Quel2ProtoAddaConfigSubsystem,
)
from quel_ic_config.quel1se_config_subsystem import Ad9082Quel1se
from quel_ic_config.quel1se_fujitsu11_config_subsystem import (
    ExstickgeCoapClientQuel1seFujitsu11,
    Quel1seFujitsu11TypeAConfigSubsystem,
    Quel1seFujitsu11TypeADebugConfigSubsystem,
    Quel1seFujitsu11TypeBConfigSubsystem,
    Quel1seFujitsu11TypeBDebugConfigSubsystem,
)
from quel_ic_config.quel1se_riken8_config_subsystem import (
    ExstickgeCoapClientQuel1seRiken8,
    ExstickgeCoapClientQuel1seRiken8WithLock,
    Quel1seRiken8ConfigSubsystem,
    Quel1seRiken8DebugConfigSubsystem,
)
from quel_ic_config.quel_clock_master_v1 import QuelClockMasterV1
from quel_ic_config.quel_config_common import QUEL1_BOXTYPE_ALIAS, Quel1BoxType, Quel1ConfigOption, Quel1Feature
from quel_ic_config.quel_ic import (
    Ad5328,
    Ad9082Generic,
    Adrf6780,
    Lmx2594,
    MixerboardGpio,
    PathselectorboardGpio,
    PowerboardPwm,
    QubeRfSwitchArray,
    Quel1TypeARfSwitchArray,
    Quel1TypeBRfSwitchArray,
)
from quel_ic_config.rfswitcharray import (
    AbstractRfSwitchArrayMixin,
    QubeRfSwitchRegs,
    QubeSwitchRegNames,
    Quel1SwitchRegNames,
    Quel1TypeARfSwitchRegs,
    Quel1TypeBRfSwitchRegs,
    RfSwitchArrayConfigHelper,
)

__version__ = "0.10.8"

__all__ = (
    "Ad5328",
    "Ad5328ConfigHelper",
    "Ad5328Regs",
    "Ad5328RegNames",
    "Ad9082Generic",
    "Ad9082Quel1",
    "Ad9082Quel1se",
    "Ad9082JesdParam",
    "Adrf6780",
    "Adrf6780ConfigHelper",
    "Adrf6780Regs",
    "Adrf6780RegNames",
    "Adrf6780LoSideband",
    "ExstickgeCoapClientAdda",
    "ExstickgeCoapClientQuel1seRiken8",
    "ExstickgeCoapClientQuel1seRiken8WithLock",
    "ExstickgeCoapClientQuel1seFujitsu11",
    "ExstickgeSockClientQuel1WithDummyLock",
    "LinkStatus",
    "Lmx2594",
    "Lmx2594ConfigHelper",
    "Lmx2594Regs",
    "Lmx2594RegNames",
    "MixerboardGpioRegNames",
    "MixerboardGpioRegs",
    "MixerboardGpioConfigHelper",
    "MixerboardGpio",
    "NcoFtw",
    "Quel1Thermistor",
    "Quel1NormalThermistor",
    "Quel1PathSelectorThermistor",
    "Quel1seOnboardThermistor",
    "Quel1seExternalThermistor",
    "ChipTemperatures",
    "Quel1AnyConfigSubsystem",
    "Quel1seAnyConfigSubsystem",
    "Quel1ConfigOption",
    "Quel1ConfigSubsystemRoot",
    "Quel1TypeARfSwitchArray",
    "Quel1TypeBRfSwitchArray",
    "Quel1TypeARfSwitchRegs",
    "Quel1TypeBRfSwitchRegs",
    "Quel1SwitchRegNames",
    "Quel1BoxType",
    "Quel1Feature",
    "QubeConfigSubsystem",
    "Quel1ConfigSubsystem",
    "QubeOuTypeAConfigSubsystem",
    "QubeOuTypeBConfigSubsystem",
    "Quel1TypeAConfigSubsystem",
    "Quel1TypeBConfigSubsystem",
    "Quel1NecConfigSubsystem",
    "Quel1seAddaConfigSubsystem",
    "Quel2ProtoAddaConfigSubsystem",
    "Quel1seRiken8ConfigSubsystem",
    "Quel1seRiken8DebugConfigSubsystem",
    "Quel1seFujitsu11TypeAConfigSubsystem",
    "Quel1seFujitsu11TypeADebugConfigSubsystem",
    "Quel1seFujitsu11TypeBConfigSubsystem",
    "Quel1seFujitsu11TypeBDebugConfigSubsystem",
    "QubeRfSwitchRegs",
    "QubeSwitchRegNames",
    "QubeRfSwitchArray",
    "QUEL1_BOXTYPE_ALIAS",
    "RfSwitchArrayConfigHelper",
    "AbstractRfSwitchArrayMixin",
    "Quel1seBoard",
    "Quel1seTempctrlState",
    "PathselectorboardGpioConfigHelper",
    "PathselectorboardGpioRegNames",
    "PathselectorboardGpioRegs",
    "PathselectorboardGpio",
    "PowerboardPwmConfigHelper",
    "PowerboardPwmRegsName",
    "PowerboardPwmRegs",
    "PowerboardPwm",
    "get_exstickge_server_info",
    "Quel1WaveSubsystem",
    "E7FwType",
    "AbstractQuel1E7ResourceMapper",
    "Quel1ConventionalE7ResourceMapper",
    "LinkupStatistic",
    "LinkupStatus",
    "LinkupFpgaMxfe",
    "AwgParam",
    "WaveChunk",
    "CapParam",
    "CapSection",
    "CapIqDataReader",
    "AbstractStartAwgunitsTask",
    "StartAwgunitsNowTask",
    "StartAwgunitsTimedTask",
    "StartCapunitsNowTask",
    "StartCapunitsByTriggerTask",
    "Quel1LineType",
    "Quel1BoxIntrinsic",
    "BoxIntrinsicStartCapunitsNowTask",
    "BoxIntrinsicStartCapunitsByTriggerTask",
    "Quel1PortType",
    "parse_port_str",
    "Quel1Box",
    "Quel1ConfigLoader",
    "BoxStartCapunitsNowTask",
    "BoxStartCapunitsByTriggerTask",
    "BoxLockError",
    "set_trancated_traceback_for_lock_error",
    "NoRfSwitchError",
    "NoLoopbackPathError",
    "force_unlock_all_boxes",
    "QuelClockMasterV1",
)



import logging
import time
from dataclasses import dataclass, field
from enum import IntEnum
from typing import Dict, Final, Set, Tuple, Union, cast

from quel_ic_config.abstract_ic import (
    AbstractIcConfigHelper,
    AbstractIcMixin,
    AbstractIcReg,
    b_1bf_bool,
    b_nbf,
    p_1bf_bool,
    p_nbf,
)

logger = logging.getLogger(__name__)


@dataclass
class Adrf6780Control(AbstractIcReg):
    parity_en: bool = field(default=False)  # [15]
    soft_reset: bool = field(default=False)  # [14]
    chip_id: int = field(default=0)  # [11:4]
    chip_rev: int = field(default=0)  # [3:0]

    def parse(self, v: int) -> None:
        self.parity_en = p_1bf_bool(v, 15)
        self.soft_reset = p_1bf_bool(v, 14)
        self.chip_id = p_nbf(v, 11, 4)
        self.chip_rev = p_nbf(v, 3, 0)

    def build(self) -> int:
        return (
            b_1bf_bool(self.parity_en, 15)
            | b_1bf_bool(self.soft_reset, 14)
            | b_nbf(self.chip_id, 11, 4)
            | b_nbf(self.chip_rev, 3, 0)
        )


@dataclass
class Adrf6780Alarm(AbstractIcReg):
    parity_error: bool = field(default=False)  # [15]
    too_few_errors: bool = field(default=False)  # [14]
    too_many_errors: bool = field(default=False)  # [13], 7@reset
    address_range_error: bool = field(default=False)  # [12], 6@reset, but may be incremented in the future.

    def parse(self, v: int) -> None:
        self.parity_error = p_1bf_bool(v, 15)
        self.too_few_errors = p_1bf_bool(v, 14)
        self.too_many_errors = p_1bf_bool(v, 13)
        self.address_range_error = p_1bf_bool(v, 12)

    def build(self) -> int:
        return (
            b_1bf_bool(self.parity_error, 15)
            | b_1bf_bool(self.too_few_errors, 14)
            | b_1bf_bool(self.too_many_errors, 13)
            | b_1bf_bool(self.address_range_error, 12)
        )


class Adrf6780AlarmReadBack(Adrf6780Alarm):
    pass


class Adrf6780AlarmMask(Adrf6780Alarm):
    pass


@dataclass
class Adrf6780Enable(AbstractIcReg):
    vga_buffer_enable: bool = field(default=False)  # [8]
    detector_enable: bool = field(default=False)  # [7]
    lo_buffer_enable: bool = field(default=False)  # [6]
    if_mode_enable: bool = field(default=False)  # [5]
    iq_mode_enable: bool = field(default=False)  # [4]
    lo_x2_enable: bool = field(default=False)  # [3]
    lo_ppf_enable: bool = field(default=False)  # [2]
    lo_enable: bool = field(default=False)  # [1]
    uc_bias_enable: bool = field(default=False)  # [0]

    def parse(self, v: int) -> None:
        self.vga_buffer_enable = p_1bf_bool(v, 8)
        self.detector_enable = p_1bf_bool(v, 7)
        self.lo_buffer_enable = p_1bf_bool(v, 6)
        self.if_mode_enable = p_1bf_bool(v, 5)
        self.iq_mode_enable = p_1bf_bool(v, 4)
        self.lo_x2_enable = p_1bf_bool(v, 3)
        self.lo_ppf_enable = p_1bf_bool(v, 2)
        self.lo_enable = p_1bf_bool(v, 1)
        self.uc_bias_enable = p_1bf_bool(v, 0)

    def build(self) -> int:
        return (
            b_1bf_bool(self.vga_buffer_enable, 8)
            | b_1bf_bool(self.detector_enable, 7)
            | b_1bf_bool(self.lo_buffer_enable, 6)
            | b_1bf_bool(self.if_mode_enable, 5)
            | b_1bf_bool(self.iq_mode_enable, 4)
            | b_1bf_bool(self.lo_x2_enable, 3)
            | b_1bf_bool(self.lo_ppf_enable, 2)
            | b_1bf_bool(self.lo_enable, 1)
            | b_1bf_bool(self.uc_bias_enable, 0)
        )


@dataclass
class Adrf6780Linearize(AbstractIcReg):
    rdac_linearize: int = field(default=0)  # [7:0]

    def parse(self, v: int) -> None:
        self.rdac_linearize = p_nbf(v, 7, 0)

    def build(self) -> int:
        return b_nbf(self.rdac_linearize, 7, 0)


class Adrf6780LoSideband(IntEnum):
    Usb = 0
    Lsb = 1


@dataclass
class Adrf6780LoPath(AbstractIcReg):
    lo_sideband: Adrf6780LoSideband = field(default=Adrf6780LoSideband.Usb)  # [10]
    q_path_phase_accuracy: int = field(default=0)  # [7:4]
    i_path_phase_accuracy: int = field(default=0)  # [3:0]

    def parse(self, v: int) -> None:
        self.lo_sideband = Adrf6780LoSideband.Lsb if p_1bf_bool(v, 10) else Adrf6780LoSideband.Usb
        self.q_path_phase_accuracy = p_nbf(v, 7, 4)
        self.i_path_phase_accuracy = p_nbf(v, 3, 0)

    def build(self) -> int:
        return (
            b_nbf(int(self.lo_sideband), 10, 10)
            | b_nbf(self.q_path_phase_accuracy, 7, 4)
            | b_nbf(self.i_path_phase_accuracy, 3, 0)
        )


@dataclass
class Adrf6780AdcControl(AbstractIcReg):
    vdet_output_select: bool = field(default=False)  # [3]
    adc_start: bool = field(default=False)  # [2]
    adc_enable: bool = field(default=False)  # [1]
    adc_clock_enable: bool = field(default=False)  # [0]

    def parse(self, v: int) -> None:
        self.vdet_output_select = p_1bf_bool(v, 3)
        self.adc_start = p_1bf_bool(v, 2)
        self.adc_enable = p_1bf_bool(v, 1)
        self.adc_clock_enable = p_1bf_bool(v, 0)

    def build(self) -> int:
        return (
            b_1bf_bool(self.vdet_output_select, 3)
            | b_1bf_bool(self.adc_start, 2)
            | b_1bf_bool(self.adc_enable, 1)
            | b_1bf_bool(self.adc_clock_enable, 0)
        )


@dataclass
class Adrf6780AdcOutput(AbstractIcReg):
    adc_status: bool = field(default=False)  # [8]
    adc_value: int = field(default=0)  # [7:0]

    def parse(self, v: int) -> None:
        self.adc_status = p_1bf_bool(v, 8)
        self.adc_value = p_nbf(v, 7, 0)

    def build(self) -> int:
        return b_1bf_bool(self.adc_status, 8) | b_nbf(self.adc_value, 7, 0)


Adrf6780Regs: Dict[int, type] = {
    0: Adrf6780Control,
    1: Adrf6780AlarmReadBack,
    2: Adrf6780AlarmMask,
    3: Adrf6780Enable,
    4: Adrf6780Linearize,
    5: Adrf6780LoPath,
    6: Adrf6780AdcControl,
    12: Adrf6780AdcOutput,
}


Adrf6780RegNames: Dict[str, int] = {
    "Control": 0,
    "AlarmReadBack": 1,
    "AlarmMask": 2,
    "Enable": 3,
    "Linearize": 4,
    "LoPath": 5,
    "AdcControl": 6,
    "AdcOutput": 12,
}


class Adrf6780Mixin(AbstractIcMixin):
    Regs = Adrf6780Regs
    RegNames = Adrf6780RegNames
    _CHIP_ID: Final[int] = 7
    _DEFAULT_EXPECTED_REVISION: Final[Set[int]] = {6}
    DEFAULT_SIDEBAND_SWITCH_WAIT = 0.001

    def __init__(self, name):
        super().__init__(name)

    def dump_regs(self) -> Dict[int, int]:
        """dumping all the available registers.

        :return: a mapping between an address and a value of the registers
        """
        regs = {}
        for addr in (0, 1, 2, 3, 4, 5, 6, 12):
            regs[addr] = self.read_reg(addr)
        return regs

    def soft_reset(self, parity_enable: bool = False) -> None:
        """committing soft reset.

        :return: None
        """
        addr, reg = cast(Tuple[int, Adrf6780Control], self._read_and_parse_reg("Control"))
        reg.parity_en = parity_enable
        reg.soft_reset = True
        self._build_and_write_reg(addr, reg)
        reg.soft_reset = False
        self._build_and_write_reg(addr, reg)

    def check_id(self, expected_revision: Union[Set[int], None] = None) -> Tuple[bool, Tuple[int, int]]:
        """confirming the chip version and chip revision, or check the soundness of the register access.

        :param expected_revision: a set of valid revisions.
        :return: an actual chip revision
        """
        if expected_revision is None:
            expected_revision = self._DEFAULT_EXPECTED_REVISION
        addr, reg = cast(Tuple[int, Adrf6780Control], self._read_and_parse_reg("Control"))
        if reg.chip_id == self._CHIP_ID and reg.chip_rev in expected_revision:
            return True, (reg.chip_id, reg.chip_rev)
        else:
            return False, (reg.chip_id, reg.chip_rev)

    def read_alarm(self) -> Dict[str, bool]:
        # don't use helper to confirm type safety.
        _, reg = cast(Tuple[int, Adrf6780AlarmReadBack], self._read_and_parse_reg("AlarmReadBack"))
        return {
            "parity": reg.parity_error,
            "too_few": reg.too_few_errors,
            "too_many": reg.too_many_errors,
            "address_range": reg.address_range_error,
        }

    def get_lo_sideband(self) -> Adrf6780LoSideband:
        """setting LO Sideband to either of LSB or USB.

        :return: None
        """
        # don't use helper to confirm type safety.
        _, reg = self._read_and_parse_reg("LoPath")
        return cast(Adrf6780LoPath, reg).lo_sideband

    def set_lo_sideband(self, sideband: Adrf6780LoSideband, wait=DEFAULT_SIDEBAND_SWITCH_WAIT) -> None:
        """setting LO Sideband to either of LSB or USB.

        :param sideband: either of LoSideband.Lsb or LoSideband.Usb.
        :return: None
        """
        # don't use helper to confirm type safety.
        if sideband != Adrf6780LoSideband.Lsb and sideband != Adrf6780LoSideband.Usb:
            raise ValueError(f"invalid sideband specifier {sideband:d} is given")

        addr, reg = self._read_and_parse_reg("LoPath")
        cast(Adrf6780LoPath, reg).lo_sideband = sideband
        self.write_reg(addr, reg.build())
        time.sleep(wait)

    def enable_detector_adc(self):
        # don't use helper to confirm type safety.
        addr_enable, reg_enable = self._read_and_parse_reg("Enable")
        cast(Adrf6780Enable, reg_enable).detector_enable = True
        self.write_reg(addr_enable, reg_enable.build())

        addr_adc_control, reg_adc_control = cast(Tuple[int, Adrf6780AdcControl], self._read_and_parse_reg("AdcControl"))
        reg_adc_control.adc_clock_enable = True
        self._build_and_write_reg(addr_adc_control, reg_adc_control)
        reg_adc_control.adc_enable = True
        self._build_and_write_reg(addr_adc_control, reg_adc_control)

    def disable_detector_adc(self):
        # don't use helper to confirm type safety.
        addr_adc_control, reg_adc_control = cast(Tuple[int, Adrf6780AdcControl], self._read_and_parse_reg("AdcControl"))
        reg_adc_control.adc_clock_enable = False
        reg_adc_control.adc_enable = False
        reg_adc_control.adc_start = False
        self._build_and_write_reg(addr_adc_control, reg_adc_control)

        addr_enable, reg_enable = cast(Tuple[int, Adrf6780Enable], self._read_and_parse_reg("Enable"))
        reg_enable.detector_enable = False
        self._build_and_write_reg(addr_enable, reg_enable)

    def read_detector_adc(self) -> int:
        """reading the value of the internal ADC. This method assumes the ADC is already enabled with the steps 1 -- 3
        described at the p.23 of the revision D datasheet.

        :return: the reading of the ADC (8bit unsigned).
        """
        # don't use helper to confirm type safety.
        addr_adc_control, reg_adc_control = cast(Tuple[int, Adrf6780AdcControl], self._read_and_parse_reg("AdcControl"))
        addr_adc_output, reg_adc_output = cast(Tuple[int, Adrf6780AdcOutput], self._read_and_parse_reg("AdcOutput"))

        reg_adc_control.adc_clock_enable = True
        reg_adc_control.adc_enable = True
        reg_adc_control.adc_start = True
        self._build_and_write_reg(addr_adc_control, reg_adc_control)
        time.sleep(200e-6)
        for i in range(10):
            reg_adc_output.parse(self.read_reg(addr_adc_output))
            if reg_adc_output.adc_status:
                break
            time.sleep(20e-6)
        else:
            raise RuntimeError(f"Timeout for ADC_STATUS of {self.name}")

        reg_adc_control.adc_start = False
        self._build_and_write_reg(addr_adc_control, reg_adc_control)
        reg_adc_output.parse(self.read_reg(addr_adc_output))
        return reg_adc_output.adc_value


class Adrf6780ConfigHelper(AbstractIcConfigHelper):
    """Helper class for programming ADRF6780 with convenient notations. It also provides caching capability that
    keep modifications on the registers to write them at once with flash_updated() in a right order.
    """

    def __init__(self, ic: Adrf6780Mixin):
        super().__init__(ic)

    def flush(self, discard_after_flush=True):
        """updating the modified registers in the order of addresses 0, 2, 3, 4, 5, 6. The modifications on addresses 1
        and 12 are ignored because they are read-only registers."""
        for addr in (0, 2, 3, 4, 5, 6):
            if addr in self.updated:
                self.ic.write_reg(addr, self.updated[addr])
        if discard_after_flush:
            self.discard()



from argparse import ArgumentParser, Namespace
from ipaddress import ip_address
from pathlib import Path
from typing import Collection, List, Set, Tuple, Union

from quel_ic_config.quel_config_common import QUEL1_BOXTYPE_ALIAS, Quel1BoxType, Quel1ConfigOption


def parse_boxtype(boxtypename: str) -> Quel1BoxType:
    if boxtypename not in QUEL1_BOXTYPE_ALIAS:
        raise ValueError(f"invalid boxtype: '{boxtypename}'")
    return Quel1BoxType.fromstr(boxtypename)


def parse_config_options(optstr: str) -> List[Quel1ConfigOption]:
    return [Quel1ConfigOption(s) for s in optstr.split(",") if s != ""]


def _parse_mxfe_combination(mxfe_combination: str) -> Tuple[int, ...]:
    if mxfe_combination == "0":
        mxfe_list: Tuple[int, ...] = (0,)
    elif mxfe_combination == "1":
        mxfe_list = (1,)
    elif mxfe_combination in {"both", "0,1"}:
        mxfe_list = (0, 1)
    elif mxfe_combination == "1,0":
        mxfe_list = (1, 0)
    elif mxfe_combination == "none":
        mxfe_list = ()
    else:
        raise ValueError(f"invalid combination of MxFEs: {mxfe_combination}")
    return mxfe_list


def _parse_comma_separated_adrf6780_list(cslist: str) -> Set[int]:
    parsed = [int(x) for x in cslist.split(",")]
    invalid = [x for x in parsed if not 0 <= x <= 7]
    if len(invalid) != 0:
        raise ValueError(f"invalid indices of ADRF6780s: {','.join([str(x) for x in invalid])}")
    return set(parsed)


def _parse_comma_separated_lmx2594_list(cslist: str) -> Set[int]:
    parsed = [int(x) for x in cslist.split(",")]
    invalid = [x for x in parsed if not 0 <= x <= 9]
    if len(invalid) != 0:
        raise ValueError(f"invalid indices of LMX2594s: {','.join([str(x) for x in invalid])}")
    return set(parsed)


def complete_ipaddrs(args: Namespace):
    if int(args.ipaddr_sss) == 0:
        args.ipaddr_sss = args.ipaddr_wss + (1 << 16)
    if int(args.ipaddr_css) == 0:
        args.ipaddr_css = args.ipaddr_wss + (4 << 16)


def add_common_arguments(
    parser: ArgumentParser,
    strict_ipaddrs: bool = False,
    use_ipaddr_wss: bool = True,
    use_ipaddr_sss: bool = True,
    use_ipaddr_css: bool = True,
    use_boxtype: bool = True,
    use_config_root: bool = True,
    use_config_options: bool = True,
    use_mxfe: bool = False,
    default_boxtype: Union[str, None] = None,
    default_config_root: Union[Path, None] = None,
    default_config_options: Union[Collection[str], None] = None,
    allow_implicit_mxfe: bool = False,
    use_hard_reset: bool = False,
) -> None:
    """adding common arguments to testlibs of quel_ic_config for manual tests. allowing to accept unused arguments for
    your convenience

    :param parser:
    :param strict_ipaddrs:
    :param use_ipaddr_wss:
    :param use_ipaddr_sss:
    :param use_ipaddr_css:
    :param use_boxtype:
    :param use_config_root:
    :param use_config_options:
    :param use_mxfe:
    :param default_boxtype:
    :param default_config_root:
    :param default_config_options:
    :param allow_implicit_mxfe:
    :return:
    """
    non_existent_ipaddress = ip_address("241.3.5.6")

    if use_ipaddr_wss:
        parser.add_argument(
            "--ipaddr_wss",
            type=ip_address,
            required=True,
            help="IP address of the wave generation/capture subsystem of the target box",
        )
    else:
        parser.add_argument(
            "--ipaddr_wss", type=ip_address, required=False, default=non_existent_ipaddress, help="IGNORED"
        )

    if use_ipaddr_sss:
        if strict_ipaddrs:
            parser.add_argument(
                "--ipaddr_sss",
                type=ip_address,
                required=True,
                help="IP address of the wave sequencer subsystem of the target box",
            )
        else:
            parser.add_argument(
                "--ipaddr_sss",
                type=ip_address,
                default=ip_address(0),
                help="IP address of the wave sequencer subsystem of the target box",
            )
    else:
        parser.add_argument(
            "--ipaddr_sss", type=ip_address, required=False, default=non_existent_ipaddress, help="IGNORED"
        )

    if use_ipaddr_css:
        if strict_ipaddrs:
            parser.add_argument(
                "--ipaddr_css",
                type=ip_address,
                required=True,
                help="IP address of the configuration subsystem of the target box",
            )
        else:
            parser.add_argument(
                "--ipaddr_css",
                type=ip_address,
                default=ip_address(0),
                help="IP address of the configuration subsystem of the target box",
            )
    else:
        parser.add_argument(
            "--ipaddr_css", type=ip_address, required=False, default=non_existent_ipaddress, help="IGNORED"
        )

    if use_boxtype:
        parser.add_argument(
            "--boxtype",
            type=parse_boxtype,
            required=True,
            help=f"a type of the target box: either of "
            f"{', '.join([t for t in QUEL1_BOXTYPE_ALIAS if not t.startswith('x_')])}",
        )
    else:
        if default_boxtype is None:
            raise ValueError("default boxtype is required when boxtype parameter is disabled")
        parser.add_argument(
            "--boxtype", type=parse_boxtype, default=default_boxtype, help="default value should be used"
        )

    if use_config_root:
        parser.add_argument(
            "--config_root",
            type=Path,
            default=None,
            help="path to configuration file root",
        )
    else:
        parser.add_argument(
            "--config_root",
            action="store_const",
            const=default_config_root,
            help="constant value",
        )

    if use_config_options:
        parser.add_argument(
            "--config_options",
            type=parse_config_options,
            default=[],
            help=f"comma separated list of config options: ("
            f"{' '.join([o for o in Quel1ConfigOption if not o.startswith('x_')])})",
        )
    else:
        if default_config_options is None:
            default_config_options = []
        parser.add_argument(
            "--config_options",
            action="store_const",
            const=default_config_options,
            help="constant value",
        )

    if use_mxfe:
        if allow_implicit_mxfe:
            parser.add_argument(
                "--mxfe",
                metavar="MxFEs",
                type=_parse_mxfe_combination,
                default=(0, 1),
                help="combination of MxFEs under test, possible values are '0', '1', 'both', '0,1', '1,0', and 'none'",
            )
        else:
            parser.add_argument(
                "--mxfe",
                metavar="MxFEs",
                type=_parse_mxfe_combination,
                required=True,
                help="combination of MxFEs under test, possible values are '0', '1', 'both', '0,1', '1,0', and 'none'",
            )

    if use_hard_reset:
        parser.add_argument(
            "--hard_reset",
            action="store_true",
            default=False,
            help="enable hard reset of AD9082 at the initialization",
        )


def add_common_workaround_arguments(
    parser: ArgumentParser,
    use_ignore_crc_error_of_mxfe: bool = False,
    use_ignore_access_failure_of_adrf6780: bool = False,
    use_ignore_lock_failure_of_lmx2594: bool = False,
    use_ignore_extraordinary_converter_select_of_mxfe: bool = False,
):
    if use_ignore_crc_error_of_mxfe:
        parser.add_argument(
            "--ignore_crc_error_of_mxfe",
            metavar="MxFEs",
            type=_parse_mxfe_combination,
            default=(),
            help="a comma-separated list of MxFEs to ignore crc_error_flag during initialization, "
            "possible values are '0', '1', 'both', '0,1', '1,0', and 'none'",
        )

    if use_ignore_access_failure_of_adrf6780:
        parser.add_argument(
            "--ignore_access_failure_of_adrf6780",
            metavar="ADRF6780s",
            type=_parse_comma_separated_adrf6780_list,
            default={},
            help="a comma-separated list of ADRF6780s to ignore communication failures",
        )

    if use_ignore_lock_failure_of_lmx2594:
        parser.add_argument(
            "--ignore_lock_failure_of_lmx2594",
            metavar="LMX2594s",
            type=_parse_comma_separated_lmx2594_list,
            default={},
            help="a comma-separated list of LMX2594s to ignore lock failures",
        )

    if use_ignore_extraordinary_converter_select_of_mxfe:
        parser.add_argument(
            "--ignore_extraordinary_converter_select_of_mxfe",
            metavar="MxFEs",
            type=_parse_mxfe_combination,
            default=(),
            help="a comma-separated list of MxFEs to ignore the abnormally configured converter select. "
            "possible values are '0', '1', 'both', '0,1', '1,0', and 'none'",
        )



import argparse
import logging
import socket
import sys
import time
from pathlib import Path
from pprint import pprint
from typing import Dict, Tuple

from quel_ic_config import LinkupFpgaMxfe, LinkupStatus, Quel1Box, Quel1BoxIntrinsic, Quel1seTempctrlState
from quel_ic_config_utils.common_arguments import (
    add_common_arguments,
    add_common_workaround_arguments,
    complete_ipaddrs,
)

logger = logging.getLogger()


def quel1_linkup() -> None:
    logging.basicConfig(level=logging.WARNING, format="{asctime} [{levelname:.4}] {name}: {message}", style="{")
    parser = argparse.ArgumentParser(
        description="making a QuEL-1 device ready to use. "
        "be aware that all the device configurations will be reset completely."
    )
    add_common_arguments(parser, use_mxfe=True, allow_implicit_mxfe=True)
    add_common_workaround_arguments(
        parser,
        use_ignore_crc_error_of_mxfe=True,
        use_ignore_access_failure_of_adrf6780=True,
        use_ignore_lock_failure_of_lmx2594=True,
    )
    parser.add_argument(
        "--background_noise_threshold",
        type=float,
        default=None,
        help="maximum allowable background noise amplitude of the ADCs",
    )
    parser.add_argument(
        "--use_204c",
        action="store_true",
        help="enable JESD204C link calibration instead of the conventional 204B one (default)",
    )
    parser.add_argument(
        "--use_204b",
        action="store_true",
        help="dare to use the conventional JESD204B link calibration instead of the 204C one",
    )
    parser.add_argument(
        "--use_bgcal",
        action="store_true",
        help="enable background calibration of JESD204C link (default)",
    )
    parser.add_argument(
        "--nouse_bgcal",
        action="store_true",
        help="disable background calibration of JESD204C link",
    )
    parser.add_argument(
        "--restart_tempctrl",
        action="store_true",
        help="restart temperature control to update setpoints",
    )
    parser.add_argument(
        "--skip_init",
        action="store_true",
        default=False,
        help="skip initialization of ICs other than AD9082",
    )
    parser.add_argument(
        "--hard_reset_wss",
        action="store_true",
        default=False,
        help="hard reset AWG units and CAP units to clear hardware error flags, should avoid using it when unnecessary",
    )
    parser.add_argument(
        "--save_dirpath",
        type=Path,
        default=None,
        help="path to directory to save captured wave data during link-up",
    )
    parser.add_argument(
        "--verbose",
        action="store_true",
        default=False,
        help="show verbose log",
    )

    args = parser.parse_args()
    complete_ipaddrs(args)
    if args.verbose:
        logging.getLogger().setLevel(logging.INFO)

    try:
        retcode = quel1_linkup_body(args)
        sys.exit(retcode)
    except AssertionError as e:
        if args.verbose:
            logger.exception(
                "internal error, please contact customer support with how to reproduce this issue", exc_info=e
            )
        else:
            logger.error("internal error, please contact customer support with how to reproduce this issue")
    except Exception as e:
        if args.verbose:
            logger.exception(e, exc_info=e)
        else:
            logger.error(e)
        sys.exit(-1)


def quel1_linkup_body(args: argparse.Namespace) -> int:
    mxfe_list: Tuple[int, ...] = args.mxfe

    try:
        box = Quel1Box.create(
            ipaddr_wss=str(args.ipaddr_wss),
            ipaddr_sss=str(args.ipaddr_sss),
            ipaddr_css=str(args.ipaddr_css),
            boxtype=args.boxtype,
            ignore_crc_error_of_mxfe=args.ignore_crc_error_of_mxfe,
            ignore_access_failure_of_adrf6780=args.ignore_access_failure_of_adrf6780,
            ignore_lock_failure_of_lmx2594=args.ignore_lock_failure_of_lmx2594,
        )
    except socket.timeout:
        logger.error(f"cannot access to the given IP addresses {args.ipaddr_wss} / {args.ipaddr_css}")
        return -1

    use_204b: bool = False
    if args.use_204b:
        if args.use_204c:
            raise ValueError("it is not allowed to specify both --use_204b and --use_204c at the same time")
        else:
            use_204b = True

    use_bgcal: bool = True
    if args.nouse_bgcal:
        if args.use_bgcal:
            raise ValueError("it is not allowed to specify both --use_bgcal and --nouse_bgcal at the same time")
        else:
            use_bgcal = False

    cli_retcode: int = 0
    linkup_ok: Dict[int, bool] = box.relinkup(
        mxfes_to_linkup=mxfe_list,
        config_root=args.config_root,
        config_options=args.config_options,
        hard_reset=args.boxtype.is_quel1se(),
        use_204b=use_204b,
        use_bg_cal=use_bgcal,
        restart_tempctrl=args.restart_tempctrl,
        skip_init=args.skip_init,
        hard_reset_wss=args.hard_reset_wss,
        background_noise_threshold=args.background_noise_threshold,
    )

    for mxfe_idx in mxfe_list:
        if linkup_ok[mxfe_idx]:
            print(f"ad9082-#{mxfe_idx} linked up successfully")
        else:
            logger.error(f"ad9082-#{mxfe_idx} failed to link up")
            cli_retcode = -1

    return cli_retcode


def quel1_linkstatus() -> None:
    logging.basicConfig(level=logging.WARNING, format="{asctime} [{levelname:.4}] {name}: {message}", style="{")
    parser = argparse.ArgumentParser(description="show the link status of AD9082s")
    add_common_arguments(
        parser, use_mxfe=True, allow_implicit_mxfe=True, use_config_root=False, use_config_options=False
    )
    add_common_workaround_arguments(parser, use_ignore_crc_error_of_mxfe=True)
    parser.add_argument(
        "--background_noise_threshold",
        type=float,
        default=None,
        help="maximum allowable background noise amplitude of the ADCs",
    )
    parser.add_argument(
        "--verbose",
        action="store_true",
        default=False,
        help="show verbose log",
    )

    args = parser.parse_args()
    complete_ipaddrs(args)
    if args.verbose:
        logging.getLogger().setLevel(logging.INFO)

    try:
        retcode = quel1_linkstatus_body(args)
        sys.exit(retcode)
    except AssertionError as e:
        if args.verbose:
            logger.exception(
                "internal error, please contact customer support with how to reproduce this issue", exc_info=e
            )
        else:
            logger.error("internal error, please contact customer support with how to reproduce this issue")
    except Exception as e:
        if args.verbose:
            logger.exception(e, exc_info=e)
        else:
            logger.error(e)
        sys.exit(-1)


def quel1_linkstatus_body(args: argparse.Namespace) -> int:
    mxfe_list: Tuple[int, ...] = args.mxfe

    try:
        box = Quel1BoxIntrinsic.create(
            ipaddr_wss=str(args.ipaddr_wss),
            ipaddr_sss=str(args.ipaddr_sss),
            ipaddr_css=str(args.ipaddr_css),
            boxtype=args.boxtype,
        )
    except socket.timeout:
        logger.error(f"cannot access to the given IP addresses {args.ipaddr_wss} / {args.ipaddr_css}")
        return -1

    if not isinstance(box, Quel1BoxIntrinsic):
        raise ValueError(f"boxtype {args.boxtype} is not supported currently")

    linkup_ok: Dict[int, bool] = box.reconnect(
        background_noise_threshold=args.background_noise_threshold,
        ignore_crc_error_of_mxfe=args.ignore_crc_error_of_mxfe,
        ignore_invalid_linkstatus=True,
    )

    for mxfe_idx in mxfe_list:
        try:
            ls, _, diag = box.css.check_link_status(
                mxfe_idx, ignore_crc_error=(mxfe_idx in args.ignore_crc_error_of_mxfe)
            )
            if ls and not linkup_ok[mxfe_idx]:
                diag = f"{box.css.ipaddr_css}:AD9082-#{mxfe_idx} has badly configured tx-link {diag[diag.rfind('('):]}"
            else:
                linkup_ok[mxfe_idx] = ls
            print(diag)
        except Exception as e:
            print(f"AD9082-#{mxfe_idx}: failed to sync with the hardware due to {e}")

    return 0 if all(linkup_ok.values()) else -1


def quel1_test_linkup() -> None:
    logging.basicConfig(level=logging.WARNING, format="{asctime} [{levelname:.4}] {name}: {message}", style="{")
    parser = argparse.ArgumentParser(
        description="identifying the failed step at the establishment of JESD204B/C link between FPGA and AD9082"
    )
    add_common_arguments(parser, use_mxfe=True, use_hard_reset=True)
    add_common_workaround_arguments(
        parser,
        use_ignore_crc_error_of_mxfe=True,
        use_ignore_access_failure_of_adrf6780=True,
        use_ignore_lock_failure_of_lmx2594=True,
    )
    parser.add_argument(
        "--background_noise_threshold",
        type=float,
        default=None,
        help="maximum allowable background noise amplitude of the ADCs",
    )
    parser.add_argument(
        "--use_204c",
        action="store_true",
        help="enable JESD204C link calibration instead of the conventional 204B one (default)",
    )
    parser.add_argument(
        "--use_204b",
        action="store_true",
        help="dare to use the conventional JESD204B link calibration instead of the 204C one",
    )
    parser.add_argument(
        "--use_bgcal",
        action="store_true",
        help="enable background calibration of JESD204C link (default)",
    )
    parser.add_argument(
        "--nouse_bgcal",
        action="store_true",
        help="disable background calibration of JESD204C link",
    )
    parser.add_argument(
        "--skip_init",
        action="store_true",
        default=False,
        help="skip initialization of ICs other than AD9082",
    )
    parser.add_argument(
        "--hard_reset_wss",
        action="store_true",
        default=False,
        help="hard reset AWG units and CAP units to clear hardware error flags, should avoid using it when unnecessary",
    )
    parser.add_argument(
        "--count",
        type=int,
        default=20,
        help="count of test iterations",
    )
    parser.add_argument(
        "--verbose",
        action="store_true",
        default=False,
        help="show verbose log",
    )

    args = parser.parse_args()
    complete_ipaddrs(args)
    if args.verbose:
        logging.getLogger().setLevel(logging.INFO)

    try:
        retcode = quel1_test_linkup_body(args)
        sys.exit(retcode)
    except AssertionError as e:
        if args.verbose:
            logger.exception(
                "internal error, please contact customer support with how to reproduce this issue", exc_info=e
            )
        else:
            logger.error("internal error, please contact customer support with how to reproduce this issue")
    except Exception as e:
        if args.verbose:
            logger.exception(e, exc_info=e)
        else:
            logger.error(e)
        sys.exit(-1)


def quel1_test_linkup_body(args: argparse.Namespace) -> int:
    mxfe_list: Tuple[int, ...] = args.mxfe

    try:
        box = Quel1BoxIntrinsic.create(
            ipaddr_wss=str(args.ipaddr_wss),
            ipaddr_sss=str(args.ipaddr_sss),
            ipaddr_css=str(args.ipaddr_css),
            boxtype=args.boxtype,
            ignore_crc_error_of_mxfe=args.ignore_crc_error_of_mxfe,
            ignore_access_failure_of_adrf6780=args.ignore_access_failure_of_adrf6780,
            ignore_lock_failure_of_lmx2594=args.ignore_lock_failure_of_lmx2594,
        )
    except socket.timeout:
        logger.error(f"cannot access to the given IP addresses {args.ipaddr_wss} / {args.ipaddr_css}")
        return -1

    if not isinstance(box, Quel1BoxIntrinsic):
        raise ValueError(f"boxtype {args.boxtype} is not supported currently")

    use_204b: bool = False
    if args.use_204b:
        if args.use_204c:
            raise ValueError("it is not allowed to specify both --use_204b and --use_204c at the same time")
        else:
            use_204b = True

    use_bgcal: bool = True
    if args.nouse_bgcal:
        if args.use_bgcal:
            raise ValueError("it is not allowed to specify both --use_bgcal and --nouse_bgcal at the same time")
        else:
            use_bgcal = False

    n_success: Dict[int, int] = {}
    n_failure: Dict[int, int] = {}
    for mxfe_idx in mxfe_list:
        n_success[mxfe_idx] = 0
        n_failure[mxfe_idx] = 0

    for _ in range(args.count):
        linkup_ok: Dict[int, bool] = box.relinkup(
            mxfes_to_linkup=mxfe_list,
            config_root=args.config_root,
            config_options=args.config_options,
            hard_reset=args.hard_reset,
            use_204b=use_204b,
            use_bg_cal=use_bgcal,
            skip_init=args.skip_init,
            hard_reset_wss=args.hard_reset_wss,
            background_noise_threshold=args.background_noise_threshold,
        )
        for mxfe_idx in linkup_ok:
            if linkup_ok[mxfe_idx]:
                n_success[mxfe_idx] += 1
            else:
                n_failure[mxfe_idx] += 1

    if args.background_noise_threshold is None:
        background_noise_threshold: float = LinkupFpgaMxfe._DEFAULT_BACKGROUND_NOISE_THRESHOLD_RELINKUP
    else:
        background_noise_threshold = args.background_noise_threshold

    for mxfe_idx in mxfe_list:
        print(f"mxfe-#{mxfe_idx}:")
        print(f"    SUCCESS: {n_success[mxfe_idx]}")
        print(f"    FAILURE: {n_failure[mxfe_idx]}")

        s = [x.categorize(background_noise_threshold) for x in box.linkupper.linkup_statistics[mxfe_idx]]
        for label, v in LinkupStatus.__members__.items():
            print(f"        {label}: {s.count(v)}")

    return 0


def quel1_dump_port_config() -> None:
    logging.basicConfig(level=logging.ERROR, format="{asctime} [{levelname:.4}] {name}: {message}", style="{")
    parser = argparse.ArgumentParser(description="Dump the current configuration settings of the QuEL1 device.")
    add_common_arguments(
        parser,
        use_config_root=False,
        use_config_options=False,
    )
    parser.add_argument("--json", action="store_true", default=False, help="dump port configuration in JSON format")
    parser.add_argument(
        "--verbose",
        action="store_true",
        default=False,
        help="show verbose log",
    )

    args = parser.parse_args()
    complete_ipaddrs(args)
    if args.verbose:
        logging.getLogger().setLevel(logging.INFO)

    try:
        _quel1_dump_port_config_body(args)
        sys.exit(0)
    except AssertionError as e:
        if args.verbose:
            logger.exception(
                "internal error, please contact customer support with how to reproduce this issue", exc_info=e
            )
        else:
            logger.error("internal error, please contact customer support with how to reproduce this issue")
    except Exception as e:
        if args.verbose:
            logger.exception(e, exc_info=e)
        else:
            logger.error(e)
    sys.exit(-1)


def _quel1_dump_port_config_body(args: argparse.Namespace) -> None:
    try:
        box = Quel1Box.create(
            ipaddr_wss=str(args.ipaddr_wss),
            ipaddr_sss=str(args.ipaddr_sss),
            ipaddr_css=str(args.ipaddr_css),
            boxtype=args.boxtype,
        )
    except socket.timeout:
        raise Exception(f"cannot access to the given IP addresses {args.ipaddr_wss} / {args.ipaddr_css}")

    if not isinstance(box, Quel1Box):
        raise ValueError(f"boxtype {args.boxtype} is not supported currently")

    if not all(box.reconnect(ignore_crc_error_of_mxfe=box.css.get_all_groups()).values()):
        raise Exception("failed to syncing with the hardware, check the power and link status before retrying.")

    if args.json:
        print(box.dump_box_to_jsonstr())
    else:
        pprint(box.dump_box(), sort_dicts=False)


def quel1_firmware_version() -> None:
    logging.basicConfig(level=logging.ERROR, format="{asctime} [{levelname:.4}] {name}: {message}", style="{")
    parser = argparse.ArgumentParser(description="show the firmware version")
    add_common_arguments(parser, use_config_root=False, use_config_options=False)
    parser.add_argument(
        "--verbose",
        action="store_true",
        default=False,
        help="show verbose log",
    )

    args = parser.parse_args()
    complete_ipaddrs(args)
    if args.verbose:
        logging.getLogger().setLevel(logging.INFO)

    try:
        retcode = quel1_firmware_version_body(args)
        sys.exit(retcode)
    except AssertionError as e:
        if args.verbose:
            logger.exception(
                "internal error, please contact customer support with how to reproduce this issue", exc_info=e
            )
        else:
            logger.error("internal error, please contact customer support with how to reproduce this issue")
    except Exception as e:
        if args.verbose:
            logger.exception(e, exc_info=e)
        else:
            logger.error(e)
        sys.exit(-1)


def quel1_firmware_version_body(args: argparse.Namespace) -> int:
    try:
        box = Quel1Box.create(
            ipaddr_wss=str(args.ipaddr_wss),
            ipaddr_sss=str(args.ipaddr_sss),
            ipaddr_css=str(args.ipaddr_css),
            boxtype=args.boxtype,
        )
    except socket.timeout:
        logger.error(f"cannot access to the given IP addresses {args.ipaddr_wss} / {args.ipaddr_css}")
        return -1

    print(f"e7awghw firmware version: {box.wss.fw_version}")
    print(f"type of the firmware: {box.wss.fw_type}")
    return 0


def quel1se_tempctrl_state() -> None:
    logging.basicConfig(level=logging.WARNING, format="{asctime} [{levelname:.4}] {name}: {message}", style="{")
    parser = argparse.ArgumentParser(description="show the state of temperature control subsystem of QuEL-1 SE")
    add_common_arguments(parser, use_config_root=False, use_config_options=False)
    parser.add_argument(
        "--verbose",
        action="store_true",
        default=False,
        help="show verbose log",
    )

    args = parser.parse_args()
    complete_ipaddrs(args)
    if args.verbose:
        logging.getLogger().setLevel(logging.INFO)

    try:
        retcode = quel1se_tempctrl_state_body(args)
        time.sleep(0.1)  # Notes: waiting for releasing the lock and closing coap context
        sys.exit(retcode)
    except AssertionError as e:
        if args.verbose:
            logger.exception(
                "internal error, please contact customer support with how to reproduce this issue", exc_info=e
            )
        else:
            logger.error("internal error, please contact customer support with how to reproduce this issue")
    except Exception as e:
        if args.verbose:
            logger.exception(e, exc_info=e)
        else:
            logger.error(e)
        sys.exit(-1)


def quel1se_tempctrl_state_body(args: argparse.Namespace) -> int:
    try:
        box = Quel1Box.create(
            ipaddr_wss=str(args.ipaddr_wss),
            ipaddr_sss=str(args.ipaddr_sss),
            ipaddr_css=str(args.ipaddr_css),
            boxtype=args.boxtype,
        )
    except socket.timeout:
        logger.error(f"cannot access to the given IP addresses {args.ipaddr_wss} / {args.ipaddr_css}")
        return -1

    state: Quel1seTempctrlState = box.css.get_tempctrl_state()
    if state == Quel1seTempctrlState.INIT:
        expl: str = "temperature control is not started yet"
    elif state == Quel1seTempctrlState.PRERUN:
        count: int = box.css.get_tempctrl_state_count()
        expl = f"warming up will finish in {count*10} seconds before starting the active control"
    elif state == Quel1seTempctrlState.RUN:
        expl = "active control is working"
    else:
        expl = "developer mode"

    print(f"tempctrl state is '{state.tostr().upper()}', {expl}.")
    return 0


def quel1se_tempctrl_reset() -> None:
    logging.basicConfig(level=logging.WARNING, format="{asctime} [{levelname:.4}] {name}: {message}", style="{")
    parser = argparse.ArgumentParser(
        description="restart the temperature control after the warming-up of the specified duration"
    )
    add_common_arguments(parser, use_config_root=False, use_config_options=False)
    parser.add_argument(
        "--duration",
        type=int,
        default=1200,
        help="duration of warming-up in seconds, default value is 1200s",
    )
    parser.add_argument(
        "--verbose",
        action="store_true",
        default=False,
        help="show verbose log",
    )

    args = parser.parse_args()
    complete_ipaddrs(args)
    if args.verbose:
        logging.getLogger().setLevel(logging.INFO)

    try:
        retcode = quel1se_tempctrl_reset_body(args)
        time.sleep(0.1)  # Notes: waiting for releasing the lock and closing coap context
        sys.exit(retcode)
    except AssertionError as e:
        if args.verbose:
            logger.exception(
                "internal error, please contact customer support with how to reproduce this issue", exc_info=e
            )
        else:
            logger.error("internal error, please contact customer support with how to reproduce this issue")
    except Exception as e:
        if args.verbose:
            logger.exception(e, exc_info=e)
        else:
            logger.error(e)
        sys.exit(-1)


def quel1se_tempctrl_reset_body(args: argparse.Namespace) -> int:
    try:
        box = Quel1Box.create(
            ipaddr_wss=str(args.ipaddr_wss),
            ipaddr_sss=str(args.ipaddr_sss),
            ipaddr_css=str(args.ipaddr_css),
            boxtype=args.boxtype,
        )
    except socket.timeout:
        logger.error(f"cannot access to the given IP addresses {args.ipaddr_wss} / {args.ipaddr_css}")
        return -1

    # Notes: css.start_tempctrl() is defined for QuEL-1 for the convenience.
    #        so, not enough for rejecting unsupported boxtypes.
    if args.duration < 0:
        logger.error("warm-up duration must be positive")
        return -1

    count = args.duration // 10
    if count == 0:
        count = 1
        logger.warning("warmup duration is set to 10 seconds, the shortest duration")
    elif count > 0xFFFFFFFF:
        count = 0xFFFFFFFF
        logger.warning(f"warmup duration is set to too long, clipped to {count*10} seconds, the longest duration")

    box.css.start_tempctrl(new_count=count)
    print(f"the active temperature control starts after {count*10}-sencod warming-up")
    return 0



from enum import Enum
from typing import Dict, List, Sequence, Tuple, Union

import aiocoap
import aiocoap.error

from quel_ic_config.exstickge_coap_client import Quel1seBoard, _ExstickgeCoapClientBase


class Quel1seTempctrlState(str, Enum):
    INIT = "init"
    PRERUN = "prerun"
    RUN = "run"
    DRYRUN = "dryrun"
    IDLE = "idle"

    @classmethod
    def fromstr(self, label: str) -> "Quel1seTempctrlState":
        return Quel1seTempctrlState(label)

    def tostr(self) -> str:
        return str(self.value)


class _ExstickgeCoapClientQuel1seTempctrlBase(_ExstickgeCoapClientBase):
    _AVAILABLE_BOARDS: Tuple[Quel1seBoard, ...]
    _TEMPCTRL_AD7490_NAME: Tuple[str, ...]

    def read_tempctrl_loop_count(self) -> int:
        uri = f"coap://{self._target[0]}/tempctrl/loop/count"
        res = self._core.request_and_wait(code=aiocoap.GET, uri=uri)
        self._coap_return_check(res, uri)
        return int(res.payload.decode(), 16)

    def read_tempctrl_loop_interval(self) -> float:
        uri = f"coap://{self._target[0]}/tempctrl/loop/interval"
        res = self._core.request_and_wait(code=aiocoap.GET, uri=uri)
        self._coap_return_check(res, uri)
        return int(res.payload.decode(), 16) / 1000.0  # Notes: ms --> s

    def read_tempctrl_state(self) -> Quel1seTempctrlState:
        uri = f"coap://{self._target[0]}/tempctrl/state"
        res = self._core.request_and_wait(code=aiocoap.GET, uri=uri)
        self._coap_return_check(res, uri)
        return Quel1seTempctrlState(res.payload.decode())

    def write_tempctrl_state(self, new_state: Quel1seTempctrlState) -> None:
        if not isinstance(new_state, Quel1seTempctrlState):
            raise TypeError(f"invalid tempctrl state: '{new_state}'")
        uri = f"coap://{self._target[0]}/tempctrl/state"
        res = self._core.request_and_wait(code=aiocoap.PUT, payload=new_state.value.encode(), uri=uri)
        self._coap_return_check(res, uri)

    def read_tempctrl_state_count(self) -> int:
        uri = f"coap://{self._target[0]}/tempctrl/state/count"
        res = self._core.request_and_wait(code=aiocoap.GET, uri=uri)
        self._coap_return_check(res, uri)
        return int(res.payload.decode(), 16)

    def write_tempctrl_state_count(self, new_count: int) -> None:
        if not 1 <= new_count <= 0xFFFFFFFF:
            raise ValueError("invalid count value: {new_count}")
        uri = f"coap://{self._target[0]}/tempctrl/state/count"
        res = self._core.request_and_wait(code=aiocoap.PUT, payload=f"{new_count:08x}".encode(), uri=uri)
        self._coap_return_check(res, uri)
        return

    def read_tempctrl_skip_count(self) -> int:
        uri = f"coap://{self._target[0]}/tempctrl/skip/count"
        res = self._core.request_and_wait(code=aiocoap.GET, uri=uri)
        self._coap_return_check(res, uri)
        return int(res.payload.decode(), 16)

    def read_tempctrl_error_count(self) -> int:
        uri = f"coap://{self._target[0]}/tempctrl/error/count"
        res = self._core.request_and_wait(code=aiocoap.GET, uri=uri)
        self._coap_return_check(res, uri)
        return int(res.payload.decode(), 16)

    def read_tempctrl_timeout_count(self) -> int:
        uri = f"coap://{self._target[0]}/tempctrl/timeout/count"
        res = self._core.request_and_wait(code=aiocoap.GET, uri=uri)
        self._coap_return_check(res, uri)
        return int(res.payload.decode(), 16)

    def read_tempctrl_broken_count(self) -> int:
        uri = f"coap://{self._target[0]}/tempctrl/broken/count"
        res = self._core.request_and_wait(code=aiocoap.GET, uri=uri)
        self._coap_return_check(res, uri)
        return int(res.payload.decode(), 16)

    def read_tempctrl_mismatch_count(self) -> int:
        uri = f"coap://{self._target[0]}/tempctrl/mismatch/count"
        res = self._core.request_and_wait(code=aiocoap.GET, uri=uri)
        self._coap_return_check(res, uri)
        return int(res.payload.decode(), 16)

    def read_tempctrl_mismatch_max(self) -> int:
        uri = f"coap://{self._target[0]}/tempctrl/mismatch/max"
        res = self._core.request_and_wait(code=aiocoap.GET, uri=uri)
        self._coap_return_check(res, uri)
        return int(res.payload.decode(), 16)

    def read_tempctrl_rawdata_ad7490(self, ad7490_idx: int) -> Tuple[int, List[int]]:
        ad7490_id = self._TEMPCTRL_AD7490_NAME[ad7490_idx]
        uri = f"coap://{self._target[0]}/tempctrl/rawdata/ad7490/{ad7490_id}"
        res = self._core.request_and_wait(code=aiocoap.GET, uri=uri)
        self._coap_return_check(res, uri)
        payload = res.payload.decode()
        separated = payload.split(",")
        if len(separated) != 19:
            raise RuntimeError(f"unexpected payload '{payload}' from '{uri}'")

        timestamp = int(separated[0], 16)
        validity = int(separated[1], 16)
        num_channel = int(separated[2], 16)
        if validity != 1:
            raise RuntimeError(f"failed data acquisition from '{uri}'")

        data = []
        for i in range(num_channel):
            v = int(separated[3 + i], 16)
            if (v >> 12) & 0x0F != i:
                raise RuntimeError(f"broken readings detected at {i}-th data from '{uri}'")
            data.append(v & 0xFFF)

        return timestamp, data

    def _read_tempctrl_rawdata_pwrpwm(self, label: str) -> Dict[str, List[int]]:
        if label not in {"current", "next"}:
            raise ValueError(f"unexpected label: {label}")

        uri = f"coap://{self._target[0]}/tempctrl/rawdata/pwrpwm/{label}"
        res = self._core.request_and_wait(code=aiocoap.GET, uri=uri)
        self._coap_return_check(res, uri)
        payload = res.payload.decode()
        separated = payload.split(",")
        if len(separated) != 42:
            raise RuntimeError(f"unexpected payload '{payload}' from '{uri}'")
        return {
            "fan": [int(x, 16) for x in separated[0:2]],
            "heater": [int(x, 16) for x in separated[2:42]],
        }

    def read_tempctrl_rawdata_pwrpwm_current(self) -> Dict[str, List[int]]:
        return self._read_tempctrl_rawdata_pwrpwm("current")

    def read_tempctrl_rawdata_pwrpwm_next(self) -> Dict[str, List[int]]:
        return self._read_tempctrl_rawdata_pwrpwm("next")

    def write_tempctrl_rawdata_pwrpwm_next(self, fan: Sequence[int], heater: Sequence[int]) -> None:
        if len(fan) != 2:
            raise ValueError("number of fan must be 2")
        if len(heater) != 40:
            raise ValueError("number of heater must be 40")

        for i, x in enumerate(fan):
            if not (isinstance(x, int) and 0 <= x <= 1000):
                raise ValueError(f"invalid data for fan[{i}]: '{x}'")
        for i, x in enumerate(heater):
            if not (isinstance(x, int) and 0 <= x <= 1000):
                raise ValueError(f"invalid data for heater[{i}]: '{x}'")

        uri = f"coap://{self._target[0]}/tempctrl/rawdata/pwrpwm/next"
        converted = [f"{x:04x}" for x in fan]
        converted.extend([f"{x:04x}" for x in heater])
        payload = ",".join(converted)
        res = self._core.request_and_wait(code=aiocoap.PUT, payload=payload.encode(), uri=uri)
        self._coap_return_check(res, uri)

    def read_tempctrl_feedback_setpoint(self) -> Dict[str, List[float]]:
        uri = f"coap://{self._target[0]}/tempctrl/feedback/setpoint"
        res = self._core.request_and_wait(code=aiocoap.GET, uri=uri)
        self._coap_return_check(res, uri)

        payload = res.payload.decode()
        separated = payload.split(",")
        if len(separated) != 42:
            raise RuntimeError(f"unexpected payload '{payload}' from '{uri}'")

        return {
            "fan": [int(v, 16) / 256.0 for v in separated[0:2]],
            "heater": [int(v, 16) / 256.0 for v in separated[2:42]],
        }

    def write_tempctrl_feedback_setpoint(self, fan: Sequence[float], heater: Sequence[float]) -> None:
        if len(fan) != 2:
            raise ValueError("number of fan must be 2")
        if len(heater) != 40:
            raise ValueError("number of heater must be 40")

        for i, x in enumerate(fan):
            if not (isinstance(x, float) and 0.0 <= x <= 256.0):
                raise ValueError(f"invalid data for fan[{i}]: '{x}'")
        for i, x in enumerate(heater):
            if not (isinstance(x, float) and 0.0 <= x <= 256.0):
                raise ValueError(f"invalid data for heater[{i}]: '{x}'")

        uri = f"coap://{self._target[0]}/tempctrl/feedback/setpoint"
        converted = [f"{round(x * 256.0):04x}" for x in fan]
        converted.extend([f"{round(x * 256.0):04x}" for x in heater])
        payload = ",".join(converted)
        res = self._core.request_and_wait(code=aiocoap.PUT, payload=payload.encode(), uri=uri)
        self._coap_return_check(res, uri)

    def read_tempctrl_feedback_coeffcient(self) -> Dict[str, Dict[str, float]]:
        retval: Dict[str, Dict[str, float]] = {}

        for atype in ("fan", "heater"):
            uri = f"coap://{self._target[0]}/tempctrl/feedback/coeff/{atype}"
            res = self._core.request_and_wait(code=aiocoap.GET, uri=uri)
            self._coap_return_check(res, uri)

            payload = res.payload.decode()
            separated = payload.split(",")
            if len(separated) != 2:
                raise RuntimeError(f"unexpected payload '{payload}' from '{uri}'")
            retval[atype] = {"Kp": int(separated[0], 16) / 16777216.0, "Ki": int(separated[1], 16) / 16777216.0}

        return retval

    def _write_tempctrl_feedback_coefficient_sub(self, atype: str, coeff: Dict[str, float]) -> None:
        if "Kp" not in coeff:
            raise ValueError(f"No 'Kp' exists in the given coefficient for {atype}")
        if "Ki" not in coeff:
            raise ValueError(f"No 'Ki' exists in the given coefficient for {atype}")

        for k in ("Kp", "Ki"):
            if not (isinstance(coeff[k], float) and 0.0 <= coeff[k] < 1.0):
                raise ValueError(f"invalid value {coeff[k]} for coefficient '{k}' of {atype}")

        uri = f"coap://{self._target[0]}/tempctrl/feedback/coeff/{atype}"
        converted = [f"{round(coeff[k] * 16777216.0):06x}" for k in ("Kp", "Ki")]
        payload = ",".join(converted)
        res = self._core.request_and_wait(code=aiocoap.PUT, payload=payload.encode(), uri=uri)
        self._coap_return_check(res, uri)

    def write_tempctrl_feedback_coeffcient(
        self, fan: Union[Dict[str, float], None], heater: Union[Dict[str, float], None]
    ) -> None:
        if fan is not None:
            self._write_tempctrl_feedback_coefficient_sub("fan", fan)
        if heater is not None:
            self._write_tempctrl_feedback_coefficient_sub("heater", heater)

    def read_tempctrl_feedback_integral(self) -> Dict[str, List[float]]:
        uri = f"coap://{self._target[0]}/tempctrl/feedback/integral"
        res = self._core.request_and_wait(code=aiocoap.GET, uri=uri)
        self._coap_return_check(res, uri)

        payload = res.payload.decode()
        separated = payload.split(",")
        if len(separated) != 42:
            raise RuntimeError(f"unexpected payload '{payload}' from '{uri}'")

        return {
            "fan": [int(v, 16) / 16777216.0 for v in separated[0:2]],
            "heater": [int(v, 16) / 16777216.0 for v in separated[2:42]],
        }

    def read_tempctrl_test_ad7490(self, ad7490_idx: int) -> Tuple[int, int, int, List[int]]:
        ad7490_id = self._TEMPCTRL_AD7490_NAME[ad7490_idx]
        uri = f"coap://{self._target[0]}/tempctrl/test/rawdata/ad7490/{ad7490_id}"
        res = self._core.request_and_wait(code=aiocoap.GET, uri=uri)
        self._coap_return_check(res, uri)
        payload = res.payload.decode()
        separated = payload.split(",")
        if len(separated) != 19:
            raise RuntimeError(f"unexpected payload '{payload}' from '{uri}'")

        timestamp = int(separated[0], 16)
        validity = int(separated[1], 16)
        num_channel = int(separated[2], 16)  # Notes: must be 16, actually.
        data = [int(separated[3 + i], 16) for i in range(num_channel)]
        return timestamp, validity, num_channel, data



#!/usr/bin/env python3

import argparse
import logging
import sys
import time
from pathlib import Path
from typing import Final

LOCK_DIR: Final[Path] = Path("/run/quelware")
TIMEOUT_DEFAULT: Final[int] = 30

logger = logging.getLogger(__name__)


def list_lockfiles():
    """
    Lists lock files (files that do not contain '|').
    """
    if not LOCK_DIR.is_dir():
        raise FileNotFoundError(f"Lock directory not found: {LOCK_DIR}")

    for f in LOCK_DIR.iterdir():
        if f.is_file() and "|" not in f.name:
            print(f.name)


def kill(lockfile_path: Path, timeout: int = TIMEOUT_DEFAULT):
    """
    Sends a request to release the specified lock (by creating a .kill file)
    and waits for it to be released.
    """

    if not lockfile_path.is_file():
        raise FileNotFoundError(f"Lock file not found: {lockfile_path}")

    try:
        claimfile: str = lockfile_path.read_text().strip()
        if not claimfile:
            raise ValueError(f"Lock file is empty: {lockfile_path}")
        if not claimfile.startswith(str(LOCK_DIR)):
            raise ValueError(f"Invalid content in lock file: '{claimfile}'")
        if not Path(claimfile).exists():
            raise ValueError(f"Invalid content in lock file: '{claimfile}'")
    except IOError as e:
        raise ValueError(f"Error reading lock file {lockfile_path}: {e}")

    killfile_path = Path(claimfile + ".kill")

    try:
        try:
            killfile_path.touch()
        except IOError as e:
            raise RuntimeError(f"Could not create kill file: {killfile_path}, reason: {e}")

        start_time = time.time()
        is_removed = False

        logger.info(f"Waiting for lock file '{lockfile_path}' to be removed (timeout: {timeout}s)...")

        while time.time() - start_time < timeout:
            if not lockfile_path.exists():
                is_removed = True
                break
            time.sleep(0.4)

        if is_removed:
            logger.info(f"Successfully confirmed removal of lock: {lockfile_path}")
        else:
            logger.info("Failed to kill the lock, but the specified lock file may no longer be active.")
            raise TimeoutError("Timeout occurred while waiting for the lock file to be removed.")

    finally:
        killfile_path.unlink(missing_ok=True)


def main():
    parser = argparse.ArgumentParser(
        prog="quel_filelock_manager", description="A command-line tool to manage Quelware file locks."
    )

    logging.basicConfig(level=logging.INFO, format="{asctime} [{levelname:.4}] {name}: {message}", style="{")

    subparsers = parser.add_subparsers(title="Commands", dest="command", metavar="COMMAND")
    subparsers.required = True

    parser_list = subparsers.add_parser("list", help="List active lock files.")
    parser_list.set_defaults(func=lambda _: list_lockfiles())

    parser_kill = subparsers.add_parser("kill", help="Request to kill a thread holding a lock.")
    parser_kill.add_argument("lockfile", help="The name of the lock file to target (e.g., 10.5.0.167).")
    parser_kill.add_argument(
        "-t",
        "--timeout",
        type=int,
        default=TIMEOUT_DEFAULT,
        help=f"Seconds to wait for the lock file to be removed (default: {TIMEOUT_DEFAULT}).",
    )

    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(0)

    def handle_kill(args):
        lockfile_path: Path = LOCK_DIR / args.lockfile
        timeout = args.timeout
        kill(lockfile_path, timeout)

    parser_kill.set_defaults(func=handle_kill)

    try:
        args = parser.parse_args()
        args.func(args)
    except Exception as e:
        logger.error(e)
        sys.exit(1)


if __name__ == "__main__":
    main()



import concurrent.futures
import logging
import os
from collections.abc import Collection, Iterable, Iterator
from ipaddress import IPv4Address
from typing import Any, Optional

import yaml
from pydantic import BaseModel, Field, model_validator

import quel_ic_config as qi

logger = logging.getLogger(__name__)


class ClockMaster(BaseModel):
    ipaddr: IPv4Address


class Box(BaseModel):
    name: str
    ipaddr: IPv4Address
    boxtype: str
    ignore_crc_error_of_mxfe: set[int] = Field(default=set())
    ignore_access_failure_of_adrf6780: set[int] = Field(default=set())
    ignore_lock_failure_of_lmx2594: set[int] = Field(default=set())
    marks: set[str] = Field(default=set())


class SystemConfiguration(BaseModel):
    version: int
    clockmaster: list[ClockMaster]
    boxes: list[Box]

    @model_validator(mode="after")
    def has_unique_names(self):
        names = set()
        for b in self.boxes:
            if b.name in names:
                raise ValueError(f"All names must be unique: {b.name}")
            else:
                names.add(b.name)
        return self

    @model_validator(mode="after")
    def has_unique_ipaddrs(self):
        ipaddrs = set()
        for b in self.boxes:
            if b.ipaddr in ipaddrs:
                raise ValueError(f"All IP addresses must be unique: {b.ipaddr}")
            else:
                ipaddrs.add(b.ipaddr)
        return self

    @model_validator(mode="after")
    def has_zero_or_one_clockmaster(self):
        if len(self.clockmaster) > 1:
            raise ValueError(f"Number of clockmaster must be equal to zero or one. Found {len(self.clockmaster)}.")
        return self


def get_boxes(conf_boxes: Iterable[Box]) -> Iterator[qi.Quel1Box]:
    boxes: list[qi.Quel1Box] = []
    for b in conf_boxes:
        logger.info(f"Starting to create a box instance for {b.name}.")
        boxtype = qi.Quel1BoxType.fromstr(b.boxtype)
        box = qi.Quel1Box.create(
            ipaddr_wss=str(b.ipaddr),
            boxtype=boxtype,
            name=b.name,
            ignore_crc_error_of_mxfe=b.ignore_crc_error_of_mxfe,
            ignore_access_failure_of_adrf6780=b.ignore_access_failure_of_adrf6780,
            ignore_lock_failure_of_lmx2594=b.ignore_lock_failure_of_lmx2594,
        )
        logger.info(f"Finished to create a box instance for {b.name} successfully.")
        boxes.append(box)

    return (b for b in boxes)


def get_boxes_in_parallel(conf_boxes: Iterable[Box]) -> Iterator[qi.Quel1Box]:
    boxes: list[qi.Quel1Box] = []
    _conf_boxes = list(conf_boxes)
    if len(_conf_boxes) == 0:
        return iter([])
    with concurrent.futures.ThreadPoolExecutor(max_workers=len(_conf_boxes)) as executer:
        future_to_conf_box = {executer.submit(get_boxes, [conf_box]): conf_box for conf_box in _conf_boxes}
        for future in concurrent.futures.as_completed(future_to_conf_box):
            try:
                boxes.extend(b for b in future.result())
            except Exception as exc:
                conf_box = future_to_conf_box[future]
                logger.warning(f"A thread to connect to {conf_box.name} generates an exception: {exc}")

    return (b for b in boxes)


def reconnect_and_get_link_status(
    box: qi.Quel1Box,
    background_noise_threshold: Optional[float] = None,
    ignore_crc_error_of_mxfe: Optional[Collection[int]] = None,
) -> dict[Any, bool]:
    logger.info(f"Starting to reconnect to {box.name}.")
    status = box.reconnect(
        background_noise_threshold=background_noise_threshold,
        ignore_crc_error_of_mxfe=ignore_crc_error_of_mxfe,
        ignore_invalid_linkstatus=True,
    )
    logger.info(f"Finishing to reconnect to {box.name}.")
    logger.debug(f"link status of {box.name} is {status}.")
    return status


def reconnect_and_get_link_status_in_parallel(
    boxes: Iterable[qi.Quel1Box],
    background_noise_threshold: Optional[float] = None,
    ignore_all_crc_errors=False,
) -> dict[str, dict[Any, bool]]:
    name_to_status: dict[str, dict[Any, bool]] = {}
    _boxes = list(boxes)
    with concurrent.futures.ThreadPoolExecutor(max_workers=len(_boxes)) as executer:
        future_to_box = {
            executer.submit(
                reconnect_and_get_link_status,
                box,
                background_noise_threshold,
                (0, 1) if ignore_all_crc_errors else None,
            ): box
            for box in _boxes
        }
        for future in concurrent.futures.as_completed(future_to_box):
            box = future_to_box[future]
            try:
                name_to_status[box.name] = future.result()
            except Exception as exc:
                logger.warning(f"A thread to reconnect to {box.name} generates an exception: {exc}")

    return name_to_status


def get_clockmaster(
    conf_clockmaster: list[ClockMaster], boxes: Collection[qi.Quel1Box]
) -> Optional[qi.QuelClockMasterV1]:
    for cm in conf_clockmaster:
        clockmaster = qi.QuelClockMasterV1(ipaddr=str(cm.ipaddr), boxes=boxes)
        if not clockmaster.check_availability():
            raise Exception("Clockmaster is not available.")
        return clockmaster
    return None


def load_default_configuration() -> SystemConfiguration:
    xdg_config_home = os.environ.get("XDG_CONFIG_HOME", os.path.join(os.path.expanduser("~"), ".config"))

    config_dir = os.path.join(xdg_config_home, "quelware")
    config_file = os.path.join(config_dir, "sysconf.yaml")

    if not os.path.exists(config_file):
        raise FileNotFoundError(f"Configuration file not found at {config_file}.")

    try:
        with open(config_file, "r") as f:
            obj = yaml.safe_load(f)
    except Exception as e:
        raise ValueError(f"Error reading {config_file}: {e}") from e

    return SystemConfiguration.model_validate(obj)



import logging
from pathlib import Path
from typing import Optional

import typer
import yaml
from rich.console import Console
from typing_extensions import Annotated

import quel_ic_config as qi
from quel_ic_config_utils import configuration

from ._snr_measurement import measure_snr

logger = logging.getLogger(__name__)
err_console = Console(stderr=True)

app = typer.Typer(add_completion=False, rich_markup_mode=None)


@app.command()
def entrypoint(
    source_box_name: Annotated[str, typer.Argument(help="Name of the Box for wave generation")],
    source_port: Annotated[str, typer.Argument(help="Port for wave generation")],
    dest_box_name: Annotated[str, typer.Argument(help="Name of the Box for signal capture")],
    dest_port: Annotated[str, typer.Argument(help="Port for signal capture")],
    sysconf_path: Annotated[
        Optional[Path],
        typer.Option("--sysconf", exists=True, file_okay=True, dir_okay=False, help="System configuration file (yaml)"),
    ] = None,
    verbose: bool = typer.Option(False, "-v", "--verbose", help="Enable verbose output."),
    log_level: Annotated[
        str,
        typer.Option(
            "--log-level",
            "-l",
            help="Set the logging level (e.g., DEBUG, INFO, WARNING, ERROR, CRITICAL)",
            case_sensitive=False,
        ),
    ] = "WARNING",
):
    """
    Measure the Signal-to-Noise Ratio using a random noise signal and show the result in dB.
    """
    logging.basicConfig(
        level=getattr(logging, log_level.upper()), format="{asctime} [{levelname:.4}] {name}: {message}", style="{"
    )

    if sysconf_path:
        with open(sysconf_path) as fp:
            sysconf_object = yaml.safe_load(fp)
        sysconf = configuration.SystemConfiguration.model_validate(sysconf_object)
    else:
        sysconf = configuration.load_default_configuration()

    if sysconf is None:
        err_console.print("Could not find the configuration file.")
        raise typer.Exit(1)

    for required_box_name in [source_box_name, dest_box_name]:
        if required_box_name not in [b.name for b in sysconf.boxes]:
            err_console.print(f"'{required_box_name}' is not found in the system configuration.")
            raise typer.Exit(1)

    name_to_box = {
        box.name: box
        for box in configuration.get_boxes_in_parallel(
            b for b in sysconf.boxes if b.name in [source_box_name, dest_box_name]
        )
    }

    name_to_status = configuration.reconnect_and_get_link_status_in_parallel(name_to_box.values())
    if not all(all(status.values()) for status in name_to_status.values()):
        err_console.print(f"Error occured during reconnecting. status={name_to_status}")
        raise typer.Exit(1)

    _trigger_port = qi.parse_port_str(source_port)
    if not name_to_box[source_box_name].is_valid_port(_trigger_port):
        err_console.print(f"Invalid port: '{_trigger_port}'")
        raise typer.Exit(1)

    _dest_port = qi.parse_port_str(dest_port)
    if not name_to_box[dest_box_name].is_valid_port(_dest_port):
        err_console.print(f"Invalid port: '{_dest_port}'")
        raise typer.Exit(1)

    snr = measure_snr(name_to_box[source_box_name], _trigger_port, name_to_box[dest_box_name], _dest_port)
    if verbose:
        print(f"SNR: {snr} dB")
    else:
        print(f"{snr}")


app()



import logging
from collections.abc import Iterable
from typing import NamedTuple

import numpy as np

logger = logging.getLogger(__name__)

_SAMPLES_PER_CLOCK = 4
_SAMPLING_RATE = 500_000_000
_SAMPLING_PERIOD_PS = int(1_000_000_000_000 / _SAMPLING_RATE)
_SYSREF_PERIOD = 2000

BoxName = str


class NameSysrefCountPair(NamedTuple):
    name: str
    sysref_count: int


def _wrap_to_range(val: int, minval: int, maxval: int):
    _length = maxval - minval
    _offset = val - minval
    return minval + _offset % _length


def propose_wait_amount_by_sysref_offset(
    name_sysref_count_pairs: Iterable[tuple[str, int]], range_warning_threshold: int = 500
) -> dict[str, int]:
    name_offset_pairs: list[tuple[BoxName, int]] = []
    for pair in name_sysref_count_pairs:
        pair = NameSysrefCountPair(*pair)
        sysref_offset = pair.sysref_count % _SYSREF_PERIOD
        name_offset_pairs.append((pair.name, sysref_offset))

    # offsets are mapped to the unit circle to find the center.
    angles: list[float] = [offset / _SYSREF_PERIOD * 2 * np.pi for _, offset in name_offset_pairs]
    x_mean = np.mean(np.cos(angles))
    y_mean = np.mean(np.sin(angles))
    angle_of_center_of_mass = np.arctan2(y_mean, x_mean)
    shifted_angles = [(theta - angle_of_center_of_mass) for theta in angles]
    shifted_counts = [
        _wrap_to_range(int(shifted_angle * _SYSREF_PERIOD / 2 / np.pi), -_SYSREF_PERIOD // 2, _SYSREF_PERIOD // 2)
        for shifted_angle in shifted_angles
    ]
    if (_range := max(shifted_counts) - min(shifted_counts)) > range_warning_threshold:
        logger.warning(f"The range of count offsets from SYSREF is too large: {_range} > {range_warning_threshold}")

    logger.info(f"shifted_counts: {shifted_counts}")
    name_to_wait = {}
    for pair, shifted in zip(name_offset_pairs, shifted_counts):
        name, _ = pair
        wait_ps = -shifted * _SAMPLES_PER_CLOCK * _SAMPLING_PERIOD_PS
        name_to_wait[name] = wait_ps

    return name_to_wait



import copy
import json
import logging
import os.path as osp
from collections.abc import Collection, Mapping, Sequence
from pathlib import Path
from typing import Any, Final, Union

from pydantic.v1.utils import deep_update

from quel_ic_config.quel_config_common import Quel1BoxType, Quel1ConfigOption, Quel1Feature

logger = logging.getLogger(__name__)


class Quel1ConfigLoader:
    DEFAULT_CONFIG_ROOTPATH: Final[Path] = Path(osp.dirname(__file__)) / "settings"

    def __init__(
        self,
        boxtype: Quel1BoxType,
        num_ic: dict[str, int],
        config_options: Collection[Quel1ConfigOption],
        features: Collection[Quel1Feature],
        config_filename: Path,
        config_rootpath: Union[Path, None] = None,
    ):
        self._boxtype: Quel1BoxType = boxtype
        self._num_ic: dict[str, int] = copy.copy(num_ic)
        self._config_options: set[Quel1ConfigOption] = set(config_options)
        self._features: set[Quel1Feature] = set(features)
        self._config_filename: Path = config_filename
        self._config_rootpath: Path = config_rootpath or self.DEFAULT_CONFIG_ROOTPATH

    def _remove_comments(self, settings: dict[str, Any]) -> dict[str, Any]:
        s1: dict[str, Any] = {}
        for k, v in settings.items():
            if not k.startswith("#"):
                if isinstance(v, dict):
                    s1[k] = self._remove_comments(v)
                else:
                    s1[k] = v
        return s1

    def _match_conditional_include(self, directive: Mapping[str, Union[str, Sequence[str]]]) -> bool:
        flag: bool = True
        file: str = ""
        for k, v in directive.items():
            if k == "file":
                if isinstance(v, str):
                    file = v
                else:
                    raise TypeError(f"invalid type of 'file': {k}")
            elif k == "boxtype":  # OR
                if isinstance(v, str):
                    v = [v]
                if not isinstance(v, list):
                    raise TypeError(f"invalid type of 'boxtype': {k}")
                if self._boxtype.value[1] not in v:
                    flag = False
            elif k == "option":  # AND
                if isinstance(v, str):
                    v = [v]
                if not isinstance(v, list):
                    raise TypeError(f"invalid type of 'option': {k}")
                for op1 in v:
                    if Quel1ConfigOption(op1) not in self._config_options:
                        flag = False
            elif k == "feature":  # AND
                if isinstance(v, str):
                    v = [v]
                if not isinstance(v, list):
                    raise TypeError(f"invalid type of 'option': {k}")
                for ft1 in v:
                    if Quel1Feature(ft1) not in self._features:
                        flag = False
            elif k == "otherwise":
                if not isinstance(v, str):
                    raise TypeError(f"invalid type of 'otherwise': {k}")
            else:
                raise ValueError(f"invalid key of conditional include: {k}")

        if file == "":
            raise ValueError(f"no file is specified in conditional include: {directive}")
        return flag

    def _include_config(
        self, directive: Union[str, Mapping[str, str], Sequence[Union[str, Mapping[str, str]]]], label_for_log: str
    ) -> tuple[dict[str, Any], set[Quel1ConfigOption]]:
        fired_options: set[Quel1ConfigOption] = set()

        if isinstance(directive, str) or isinstance(directive, dict):
            directive = [directive]

        config: dict[str, Any] = {}
        for d1 in directive:
            if isinstance(d1, str):
                with open(self._config_rootpath / d1) as f:
                    logger.info(f"basic config applied to {label_for_log}: {d1}")
                    config = deep_update(config, json.load(f))
            elif isinstance(d1, dict):
                if self._match_conditional_include(d1):
                    with open(self._config_rootpath / d1["file"]) as f:
                        logger.info(f"conditional config applied to {label_for_log}: {d1}")
                        config = deep_update(config, json.load(f))
                        if "option" in d1:
                            option = d1["option"]
                            if isinstance(option, str):
                                fired_options.add(Quel1ConfigOption(option))
                            elif isinstance(option, list):
                                fired_options.update({Quel1ConfigOption(o) for o in option})
                            else:
                                raise AssertionError
                elif "otherwise" in d1:
                    with open(self._config_rootpath / d1["otherwise"]) as f:
                        logger.info(f"'otherwise' part of conditional config applied to {label_for_log}: {d1}")
                        config = deep_update(config, json.load(f))
            else:
                raise TypeError(f"malformed template at {label_for_log}: '{d1}'")
        if "meta" in config:
            del config["meta"]
        return self._remove_comments(config), fired_options

    def load_config(self) -> dict[str, Any]:
        logger.info(f"loading configuration settings from '{self._config_rootpath / self._config_filename}'")
        logger.info(f"boxtype = {self._boxtype}")
        logger.info(f"config_options = {self._config_options}")
        fired_options: set[Quel1ConfigOption] = set()

        with open(self._config_rootpath / self._config_filename) as f:
            root: dict[str, Any] = self._remove_comments(json.load(f))

        config = copy.copy(root)
        for k0, directive0 in root.items():
            if k0 == "meta":
                pass
            elif k0 in self._num_ic.keys():
                for idx, directive1 in enumerate(directive0):
                    if idx >= self._num_ic[k0]:
                        raise ValueError(f"too many {k0.upper()}s are found")
                    config[k0][idx], fired_options_1 = self._include_config(directive1, label_for_log=f"{k0}[{idx}]")
                    fired_options.update(fired_options_1)
            else:
                raise ValueError(f"invalid name of IC: '{k0}'")

        for k1, n1 in self._num_ic.items():
            if len(config[k1]) != n1:
                raise ValueError(
                    f"lacking config, there should be {n1} instances of '{k1}', "
                    f"but actually have {len(config[k1])} ones"
                )

        for option in self._config_options:
            if option not in fired_options:
                logger.warning(f"config option '{str(option)}' is not applicable")

        return config



from enum import Enum
from pathlib import Path
from typing import Dict, Final

_DEFAULT_LOCK_DIRECTORY: Final[Path] = Path("/run/quelware")


class Quel1BoxType(Enum):
    QuBE_OU_TypeA = ("qube", "ou-type-a")
    QuBE_OU_TypeB = ("qube", "ou-type-b")
    QuBE_RIKEN_TypeA = ("qube", "riken-type-a")
    QuBE_RIKEN_TypeB = ("qube", "riken-type-b")
    QuEL1_TypeA = ("quel-1", "type-a")
    QuEL1_TypeB = ("quel-1", "type-b")
    QuEL1_NTT = ("quel-1", "ntt")  # Notes: not fully supported yet.  TODO: implement automatic tests
    QuEL1_NEC = ("quel-1", "nec")
    QuEL1SE_RIKEN8 = ("quel-1se", "riken8")
    QuEL1SE_RIKEN8DBG = ("quel-1se", "riken8dbg")  # for development purpose
    QuEL1SE_FUJITSU11_TypeA = ("quel-1se", "fujitsu11-type-a")  # not available yet
    QuEL1SE_FUJITSU11_TypeB = ("quel-1se", "fujitsu11-type-b")  # not available yet
    QuEL1SE_FUJITSU11DBG_TypeA = ("quel-1se", "fujitsu11dbg-type-a")  # for development purpose
    QuEL1SE_FUJITSU11DBG_TypeB = ("quel-1se", "fujitsu11dbg-type-b")  # for development purpose
    # Notes: the following boxtypes can be eliminated without the prior notification.
    QuEL1SE_Adda = ("quel-1se", "adda")  # Prototype configuration
    QuEL2_ProtoAdda = ("quel-2", "protoadda")  # Prototype configuration

    @classmethod
    def fromstr(cls, label: str) -> "Quel1BoxType":
        return QUEL1_BOXTYPE_ALIAS[label]

    def tostr(self):
        for a, t in QUEL1_BOXTYPE_ALIAS.items():
            if t is self:
                return a
        else:
            raise AssertionError

    def is_quel1se(self) -> bool:
        # Notes: quel-2 is regarded as being quel-1se since it is in early stage and not so different from quel-1se.
        return self.value[0] in {"quel-1se", "quel-2"}


QUEL1_BOXTYPE_ALIAS: Final[Dict[str, Quel1BoxType]] = {
    "qube-ou-a": Quel1BoxType.QuBE_OU_TypeA,
    "qube-ou-b": Quel1BoxType.QuBE_OU_TypeB,
    "qube-riken-a": Quel1BoxType.QuBE_RIKEN_TypeA,
    "qube-riken-b": Quel1BoxType.QuBE_RIKEN_TypeB,
    "quel1-a": Quel1BoxType.QuEL1_TypeA,
    "quel1-b": Quel1BoxType.QuEL1_TypeB,
    "quel1-ntt": Quel1BoxType.QuEL1_NTT,
    "quel1-nec": Quel1BoxType.QuEL1_NEC,
    "quel1se-riken8": Quel1BoxType.QuEL1SE_RIKEN8,
    "quel1se-fujitsu11-a": Quel1BoxType.QuEL1SE_FUJITSU11_TypeA,
    "quel1se-fujitsu11-b": Quel1BoxType.QuEL1SE_FUJITSU11_TypeB,
    "x-quel1se-adda": Quel1BoxType.QuEL1SE_Adda,
    "x-quel1se-riken8": Quel1BoxType.QuEL1SE_RIKEN8DBG,
    "x-quel1se-fujitsu11-a": Quel1BoxType.QuEL1SE_FUJITSU11DBG_TypeA,
    "x-quel1se-fujitsu11-b": Quel1BoxType.QuEL1SE_FUJITSU11DBG_TypeB,
    "x-quel2-protoadda": Quel1BoxType.QuEL2_ProtoAdda,
}


class Quel1Feature(Enum):
    SINGLE_ADC = "single_adc"
    BOTH_ADC = "both_adc"
    # Notes: workaround for a bad design of feedback_20231108 firmware
    BOTH_ADC_EARLY = "both_adc_early"


class Quel1ConfigOption(str, Enum):
    USE_READ_IN_MXFE0 = "use_read_in_mxfe0"  # Notes: only for SIMPLEMULTI_CLASSIC
    USE_MONITOR_IN_MXFE0 = "use_monitor_in_mxfe0"  # Notes: only for SIMPLEMULTI_CLASSIC
    REFCLK_CORRECTED_MXFE0 = "refclk_corrected_mxfe0"
    REFCLK_12GHz_FOR_MXFE0 = "refclk_12ghz_for_mxfe0"
    X_REFCLK_250MHz_FOR_MXFE0 = "x_refclk_250mhz_for_mxfe0"
    X_REFCLK_100MHz_FOR_MXFE0 = "x_refclk_100mhz_for_mxfe0"
    DAC_CNCO_1500MHz_MXFE0 = "dac_cduc_1500mhz_mxfe0"
    DAC_CNCO_2000MHz_MXFE0 = "dac_cduc_2000mhz_mxfe0"

    USE_READ_IN_MXFE1 = "use_read_in_mxfe1"  # Notes: only for SIMPLEMULTI_CLASSIC
    USE_MONITOR_IN_MXFE1 = "use_monitor_in_mxfe1"  # Notes: only for SIMPLEMULTI_CLASSIC
    REFCLK_CORRECTED_MXFE1 = "refclk_corrected_mxfe1"
    REFCLK_12GHz_FOR_MXFE1 = "refclk_12ghz_for_mxfe1"
    X_REFCLK_250MHz_FOR_MXFE1 = "x_refclk_250mhz_for_mxfe1"
    X_REFCLK_100MHz_FOR_MXFE1 = "x_refclk_100mhz_for_mxfe1"
    DAC_CNCO_1500MHz_MXFE1 = "dac_cduc_1500mhz_mxfe1"
    DAC_CNCO_2000MHz_MXFE1 = "dac_cduc_2000mhz_mxfe1"
    SE8_MXFE1_AWG1331 = "se8_mxfe1_awg1331"
    SE8_MXFE1_AWG2222 = "se8_mxfe1_awg2222"
    SE8_MXFE1_AWG3113 = "se8_mxfe1_awg3113"


class Quel1RuntimeOption(Enum):
    ALLOW_DUAL_MODULUS_NCO = 1



from abc import ABCMeta, abstractmethod


class _TfpltConv:
    """Converting relative resistor value to temperature. The formula is based on page 4 of
    https://www.vishay.com/docs/33027/tfptl.pdf.
    """

    _MinR: float = 0.702  # a.u.
    _MaxR: float = 1.595  # a.u.

    @classmethod
    def convert(cls, r: float) -> float:
        """converting a relative resistor value into temperature in degC

        :param r: a relative resitor of the thermistor
        :return: temperature at the resitor in degC
        """

        if not (cls._MinR <= r <= cls._MaxR):
            raise ValueError(f"the given relative resistor value {r:.3f} is out of range")
        return ((28.54 * r - 158.5) * r + 474.8) * r - 319.85


class Quel1Thermistor(metaclass=ABCMeta):
    def __init__(self, name: str):
        self.name = name

    @abstractmethod
    def convert(self, v: int) -> float:
        pass


# the following two classws will be moved to the right place.
class Quel1NormalThermistor(Quel1Thermistor):
    def convert(self, v: int) -> float:
        # Notes: 7303 comes from strange Vdd of 4.45V.
        return _TfpltConv.convert(v / (7307 - v))


class Quel1PathSelectorThermistor(Quel1Thermistor):
    def convert(self, v: int) -> float:
        # Notes: thermistor is read by AD7490 on power board v13.
        return _TfpltConv.convert(v / (7307 - v) * 3.0303)


class Quel1seOnboardThermistor(Quel1Thermistor):
    """330Ohm thermistor + 1kOhm resistor"""

    def convert(self, v: int) -> float:
        # TODO: check whether this is OK or not.
        # Notes: Vref = 2.5V, range = Vref
        return _TfpltConv.convert(v / (8192 - v) * 3.0303)


class Quel1seExternalThermistor(Quel1Thermistor):
    """1.00kOhm thermistor + 4.70kOhm resistor"""

    def convert(self, v: int) -> float:
        # Notes: Vref = 2.5V, range = Vref
        return _TfpltConv.convert(v / (8192 - v) * 4.7)



from collections.abc import Callable
from typing import Any, Optional, Protocol
from typing_extensions import Self

import quel_ic_config as qi


class DelayMeasurer(Protocol):
    """
    A class to measure the time delay between a generated signal and a captured signal.
    """

    def measure(
        self,
        source_init_blank_word: int = 0,
        dest_init_blank_word: int = 0,
        source_timecounter_shift: int = 0,
        dest_timecounter_shift: int = 0,
        callback: Optional[Callable[[Self, dict[str, Any]], None]] = None,
    ) -> tuple[int, list[dict[str, Any]]]: ...


class RelevantPeakNotFoundError(Exception):
    def __init__(
        self,
        *args,
        source_box_name: str = "unprovided",
        source_port: qi.Quel1PortType = -1,
        dest_box_name: str = "unprovided",
        dest_port: qi.Quel1PortType = -1,
    ):
        self.source_box_name = source_box_name
        self.source_port = source_port
        self.dest_box_name = dest_box_name
        self.dest_port = dest_port
        super().__init__(*args)
